# -*- coding=utf-8 -*-

"""Reads' identifiers container abstract base class module."""

from __future__ import annotations

from abc import abstractmethod
from typing import Generator, Generic

from revsymg.lib.index_lib import FORWARD_INT, IndexT, IndOrT, OrT
from revsymg.lib.str_lib import IdT


# ============================================================================ #
#                     READS IDENTIFIERS CONTAINER ABSTRACT                     #
# ============================================================================ #
class AbstractIdsContainer(Generic[IdT]):
    """Abstract class for reads identifiers container."""

    # -*- Identifiers -*-

    # pylint: disable=missing-param-doc
    @abstractmethod
    def id_to_indor(self, read_id: IdT,
                    orientation: OrT = FORWARD_INT) -> IndOrT:
        """Return the oriented vertex associated to the id.

        Parameters
        ----------
        read_id : IdT
            Read's identifier
        orientation : OrT, default FORWARD_INT
            Read orientation

        Returns
        -------
        IndOrT
            Oriented vertex

        Raises
        ------
        NoReadId
            If there is no read's identifier in container
        """
        raise NotImplementedError

    @abstractmethod
    def indor_to_id(self, vertex: IndOrT) -> IdT:
        """Return the id from vertices.

        Parameters
        ----------
        vertex : IndOrT
            Oriented vertex

        Returns
        -------
        IdT
            The associated read's identifier

        Raises
        ------
        NoVertex
            There is no read's identifier corresponding to given vertex
        """
        raise NotImplementedError

    # -*- Setter -*-

    @abstractmethod
    def add(self, read_id: IdT) -> IndexT:
        """Add an entry if identifier not yet recorded else corresponding index.

        Parameters
        ----------
        read_id : IdT
            Read's identifier

        Returns
        -------
        IndexT
            Associated vertex index
        """
        raise NotImplementedError

    @abstractmethod
    def pop(self, read_id: IdT) -> IndexT:
        """Remove read's identifier.

        Parameters
        ----------
        read_id : IdT
            Read's identifier

        Returns
        -------
        IndexT
            Index corresponding to read's identifier

        Raises
        ------
        NoReadId
            If there is no read's identifier in container
        """
        raise NotImplementedError

    # -*- Overloaded built-in -*-

    # DOC: special method

    @abstractmethod
    def __iter__(self) -> Generator[IdT, None, None]:
        """Iterate on reads' identifier.

        Yields
        ------
        IdT
            Reads' identifier
        """
        raise NotImplementedError

    @abstractmethod
    def __contains__(self, read_id: IdT) -> bool:
        """Return True if read's identifier in container.

        Parameters
        ----------
        read_id : IdT
            Read's identifier

        Returns
        -------
        bool
            Is read's identifier in container?
        """
        raise NotImplementedError

    @abstractmethod
    def __len__(self) -> int:
        """Return the number of recorded read's identifiers.

        Returns
        -------
        int
            Number of reads' identifiers
        """
        raise NotImplementedError



# -*- coding=utf-8 -*-

"""Global edges container class module."""

from __future__ import annotations

from typing import Any, Iterator

from revsymg.exceptions import (
    NoEdge,
    NoEdgeIndex,
    NoEdgesAttribute,
    NoVertexIndex,
    WrongAttributeType,
    _NoAttribute,
    _NoKey,
)
from revsymg.graphs._adjancy_table import _AdjsT
from revsymg.graphs.attributes import AttributesContainer
from revsymg.lib.index_lib import (
    FORWARD_INT,
    FWDOR_REVOR,
    IND,
    OR,
    ORIENT_REV,
    PRED_IND,
    REVERSE_INT,
    SUCC_IND,
    EIndOrIndT,
    EIndOrT,
    IndexT,
    IndOrIndT,
    IndOrT,
)


# ============================================================================ #
#                                EDGES CONTAINER                               #
# ============================================================================ #
class Edges():
    r"""Edges container class for both edges' orientations.

    For each overlap :math:`o \in \mathcal{O}`, there are two edges
    :math:`e \in E` and :math:`\bar{e} \in E` *i.e.* one for both orientations.

    Warnings
    --------
    You must not call this class.
    Use it only for typing or get an already existing object.
    """

    def __init__(self, _adjs: _AdjsT):
        """The Initializer."""
        #
        # The number of current edges
        #
        self._card_e: int = 0
        #
        # The last edge integer identifier
        #
        self._ind_e: IndexT = -1  # FIXME: change to be always | 2 as described
        #
        # Adjancies table
        #
        # [PREDS_IND]: Predecessors lists
        #   [v_index]: [(u_indor, e_ind) | (u_indor, v_indF) in E]
        # [SUCCS_IND]: Succesors lists
        #   [u_index]: [(v_indor, e_ind) | (u_indF, v_indor) in E]
        #
        self._adjs: _AdjsT = _adjs
        #
        # Attributes container
        #
        self._attributes: AttributesContainer = (
            AttributesContainer()
        )

    # -*- Attributes -*-

    def attr(self, edge_index: IndexT, attrname: str) -> Any:
        """Return the edge attribute `attrname`.

        Parameters
        ----------
        edge_index: IndexT
            Edge's index
        attrname: str
            Attribute name

        Returns
        -------
        Any
            Attribute value

        Raises
        ------
        NoEdgeIndex
            If edge does not exist
        NoEdgesAttribute
            If there is no attribute name like this for edges

        Note
        ----
        A removed edge still have a recorded attribute.
        """
        try:
            return self._attributes.get(edge_index, attrname)
        except _NoAttribute as exc:
            raise NoEdgesAttribute(attrname) from exc
        except _NoKey as exc:
            raise NoEdgeIndex(edge_index) from exc

    def attrs(self, edge_index: IndexT) -> Iterator[tuple[str, Any]]:
        """Iterate over all edge's attributes.

        Parameters
        ----------
        edge_index: IndexT
            Edge's index

        Yields
        ------
        str
            Attribute's name
        Any
            Attribute's value

        Raises
        ------
        NoEdgeIndex
            If there is no edge index

        Note
        ----
        A removed edge still have a recorded attribute.
        """
        try:
            yield from self._attributes.get_all(edge_index)
        except _NoKey as _no_key:
            raise NoEdgeIndex(edge_index) from _no_key

    def new_attr(self, attrname: str, default: Any):
        """Add an attribute entry.

        Parameters
        ----------
        attrname : str
            Attribute's name
        default : Any
            Default attribute value

        Note
        ----
        Before setting attribute to one edge, it is necessary to add
        an attribute entry using this method.
        """
        self._attributes.new_attr(attrname, default)

    def set_attr(self, edge_index: IndexT, attrname: str, attrvalue: Any):
        """Change the value of one attribute.

        Parameters
        ----------
        edge_index : IndexT
            Edge's index
        attrname : str
            Attribute's name
        attrvalue : Any
            Attribute's value

        Raises
        ------
        NoEdgesAttribute
            If there is no attribute nammed like this for edges
        WrongAttributeType
            Given value does not correspond to attribute's type
        NoEdgeIndex
            If there is no vertex index

        Note
        ----
        Be sure having added an attribute entry using the method
        :meth:`~revsymg.Vertices.new_attr`
        """
        try:
            self._attributes.set_attr(edge_index, attrname, attrvalue)
        except _NoAttribute as _no_attr:
            raise NoEdgesAttribute(attrname) from _no_attr
        except WrongAttributeType as wrong_attr_type:
            raise wrong_attr_type
        except _NoKey as _no_key:
            raise NoEdgeIndex(edge_index) from _no_key

    # -*- Getter -*-

    def biggest_edge_index(self) -> int:
        """Return the biggest edge index.

        Returns
        -------
        int
            The biggest edge index

        Notes
        -----
        The number of distinct edges indices is equal to
        the biggest edge index plus one.
        """
        return self._ind_e

    def eindor_to_eind(self, first_vertex: IndOrT,
                       second_vertex: IndOrT) -> Iterator[IndexT]:
        """Iterate over edge indices from oriented edge.

        Parameters
        ----------
        first_vertex : IndOrT
            First oriented vertex
        second_vertex : IndOrT
            Second oriented vertex

        Yields
        ------
        IndexT
            Edge index

        Raises
        ------
        NoVertexIndex
            One of the vertex in edge does not exist
        """
        #
        # Exceptions
        #
        if first_vertex[IND] >= len(self._adjs[PRED_IND]):
            raise NoVertexIndex(first_vertex[IND])
        if second_vertex[IND] >= len(self._adjs[PRED_IND]):
            raise NoVertexIndex(second_vertex[IND])
        #
        # Search
        #
        if first_vertex[OR] == FORWARD_INT:
            yield from (
                e_ind
                for succ, e_ind in self._adjs[SUCC_IND][first_vertex[IND]]
                if succ == second_vertex
            )
        else:
            v_rev = second_vertex[IND], ORIENT_REV[second_vertex[OR]]
            yield from (
                e_ind
                for pre_rev, e_ind in self._adjs[PRED_IND][first_vertex[IND]]
                if pre_rev == v_rev
            )

    # -*- Iteration -*-

    def preds(self, vertex: IndOrT) -> Iterator[IndOrIndT]:
        """Iterate over vertex's predecessors.

        Parameters
        ----------
        vertex: IndOrT
            Oriented vertex

        Yields
        ------
        IndOrIndT
            Predecessor oriented vertex with edge index

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        """
        if vertex[OR] == FORWARD_INT:
            try:
                yield from self._adjs[PRED_IND][vertex[IND]]
            except IndexError as exc:
                raise NoVertexIndex(vertex[IND]) from exc
        else:
            try:
                for succ_rev, e_ind in self._adjs[SUCC_IND][vertex[IND]]:
                    yield (succ_rev[IND], ORIENT_REV[succ_rev[OR]]), e_ind
            except IndexError as exc:
                raise NoVertexIndex(vertex[IND]) from exc

    def succs(self, vertex: IndOrT) -> Iterator[IndOrIndT]:
        """Iterate over vertex's successors.

        Parameters
        ----------
        vertex: IndOrT
            Oriented vertex

        Yields
        ------
        IndOrIndT
            Successor oriented vertex with edge index

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        """
        if vertex[OR] == FORWARD_INT:
            try:
                yield from self._adjs[SUCC_IND][vertex[IND]]
            except IndexError as exc:
                raise NoVertexIndex(vertex[IND]) from exc
        else:
            try:
                for pred_rev, e_ind in self._adjs[PRED_IND][vertex[IND]]:
                    yield (pred_rev[IND], ORIENT_REV[pred_rev[OR]]), e_ind
            except IndexError as exc:
                raise NoVertexIndex(vertex[IND]) from exc

    def neighbours(self, vertex: IndOrT) -> Iterator[IndOrIndT]:
        """Return the generator of `vertex`'s adjacencies vertices.

        Parameters
        ----------
        vertex: IndOrT
            Oriented vertex

        Yields
        ------
        IndOrIndT
            Oriented neighbouring vertices with edge index

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        """
        try:
            yield from self.preds(vertex)
            yield from self.succs(vertex)
        except NoVertexIndex as no_vertex_index:
            raise no_vertex_index

    # -*- Setter -*-

    def add(self, first_vertex: IndOrT, second_vertex: IndOrT) -> IndexT:
        r"""Add an edge to edges container.

        Parameters
        ----------
        first_vertex : IndOrT
            First oriented vertex
        second_vertex : IndOrT
            Second oriented vertex

        Returns
        -------
        IndexT
            Edge index

        Raises
        ------
        NoVertexIndex
            If one vertex does not exist

        Notes
        -----
        Edges container contains the two orientations of each edge *i.e.*
        :math:`\forall e \in E, \bar{e} \in E`
        """
        #
        # Exceptions
        #
        if first_vertex[IND] >= len(self._adjs[PRED_IND]):
            raise NoVertexIndex(first_vertex[IND])
        if second_vertex[IND] >= len(self._adjs[PRED_IND]):
            raise NoVertexIndex(second_vertex[IND])
        self._ind_e += 1
        #
        # Successor
        #
        if first_vertex[OR] == FORWARD_INT:  # (u_f, v_f-r)
            self._adjs[SUCC_IND][first_vertex[IND]].append(
                (second_vertex, self._ind_e),
            )
        else:  # (u_r, v_f-r)
            v_r = second_vertex[IND], ORIENT_REV[second_vertex[OR]]
            self._adjs[PRED_IND][first_vertex[IND]].append(
                (v_r, self._ind_e),
            )
        #
        # Predecessor
        #
        if second_vertex[OR] == FORWARD_INT:  # (u_f-r, v_f)
            self._adjs[PRED_IND][second_vertex[IND]].append(
                (first_vertex, self._ind_e),
            )
        else:  # (u_f-r, v_r)
            u_r = first_vertex[IND], ORIENT_REV[first_vertex[OR]]
            self._adjs[SUCC_IND][second_vertex[IND]].append(
                (u_r, self._ind_e),
            )
        self._card_e += 1
        self._attributes.add_keys(1)
        return self._ind_e

    def delete(self, first_vertex: IndOrT, second_vertex: IndOrT,
               edge_index: IndexT):
        r"""Remove the edge.

        Parameters
        ----------
        first_vertex : IndOrT
            First oriented vertex
        second_vertex : IndOrT
            Second oriented vertex
        edge_index : IndexT
            Edge's index

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        NoEdge
            If there is no recorded edge with these vertices and this index

        Note
        ----
        When deleting, the edge is replacing by the last one in adjancy list,
        and the last element of the adjancy list is removed.

        Warning
        -------
        As :class::`revsymg.graphs.Edges` has the reverse symmetry property,
        removing one orientation of an edge has the consequence of removing the
        other orientation *i.e.*
        :math:`\forall e \in E, \bar{e} \notin E \setminus \{e\}`
        """
        # DOC: do a figure
        # XXX: change edge index? -> if so, do a figure
        #
        # Exceptions
        #
        if first_vertex[IND] >= len(self._adjs[PRED_IND]):
            raise NoVertexIndex(first_vertex[IND])
        if second_vertex[IND] >= len(self._adjs[PRED_IND]):
            raise NoVertexIndex(second_vertex[IND])
        #
        # Delete u from v's predecessors
        #
        u_adj: IndOrIndT = (
            (
                first_vertex[IND],
                FWDOR_REVOR[second_vertex[OR]][first_vertex[OR]],
            ),
            edge_index,
        )
        pred_ind = FWDOR_REVOR[PRED_IND][second_vertex[OR]]
        try:
            adj_ind = self._adjs[pred_ind][second_vertex[IND]].index(u_adj)
            # by symmetry, other index method below will raise ValueError too
        except ValueError as value_error:
            raise NoEdge(
                first_vertex, second_vertex, edge_index,
            ) from value_error
        if adj_ind < len(self._adjs[pred_ind][second_vertex[IND]]) - 1:
            # move the last to the place of the deleted edge and remove the last
            self._adjs[pred_ind][second_vertex[IND]][adj_ind] = (
                self._adjs[pred_ind][second_vertex[IND]].pop()
            )
        else:
            # it is the last: so just remove it
            self._adjs[pred_ind][second_vertex[IND]].pop()
        #
        # Delete v from u's successors
        #
        v_adj: IndOrIndT = (
            (
                second_vertex[IND],
                FWDOR_REVOR[first_vertex[OR]][second_vertex[OR]],
            ),
            edge_index,
        )
        succ_ind = FWDOR_REVOR[SUCC_IND][first_vertex[OR]]
        adj_ind = self._adjs[succ_ind][first_vertex[IND]].index(v_adj)
        if adj_ind < len(self._adjs[succ_ind][first_vertex[IND]]) - 1:
            # move the last to the place of the deleted edge and remove the last
            self._adjs[succ_ind][first_vertex[IND]][adj_ind] = (
                self._adjs[succ_ind][first_vertex[IND]].pop()
            )
        else:
            # it is the last: so just remove it
            self._adjs[succ_ind][first_vertex[IND]].pop()
        self._card_e -= 1

    # -*- Overloaded built-in -*-

    def __iter__(self) -> Iterator[EIndOrIndT]:
        """Iterate on oriented edges with edge index.

        Yields
        ------
        EIndOrIndT
            Oriented edges with their edge index
        """
        for v_ind in range(len(self._adjs[SUCC_IND])):
            v = (v_ind, FORWARD_INT)
            v_rev = (v_ind, REVERSE_INT)
            for w, e_ind in self._adjs[SUCC_IND][v_ind]:
                yield v, w, e_ind
                if w[OR] == FORWARD_INT:
                    yield (w[IND], ORIENT_REV[w[OR]]), v_rev, e_ind
            for u, e_ind in self._adjs[PRED_IND][v_ind]:
                if u[OR] == REVERSE_INT:
                    yield u, v, e_ind

    def __contains__(self, edge: EIndOrT) -> bool:
        """Return True if the oriented edge is in edges container.

        Parameters
        ----------
        edge: EIndOrT
            Oriented edge

        Returns
        -------
        bool
            Oriented edges in edges container
        """
        u, v = edge
        #
        # No vertices
        #
        if u[IND] >= len(self._adjs[PRED_IND]):
            return False
        if v[IND] >= len(self._adjs[PRED_IND]):
            return False
        #
        # Search in forward's successors table
        #
        if u[OR] == FORWARD_INT:
            return any(v == succ for succ, _ in self._adjs[SUCC_IND][u[IND]])
        #
        # else in reverse's predecessors table
        #
        v_rev = v[IND], ORIENT_REV[v[OR]]
        return any(
            v_rev == pred_rev
            for pred_rev, _ in self._adjs[PRED_IND][u[IND]]
        )

    def __len__(self) -> int:
        r"""Return the number of oriented edges.

        Returns
        -------
        int
            Number of oriented edges

        Notes
        -----
        :math:`\forall e \in E, \bar{e} \in E`
        """
        return 2 * self._card_e



# -*- coding=utf-8 -*-

"""Global overlaps graph class file."""

from __future__ import annotations

from typing import Any, Optional

from revsymg.exceptions import NoGraphAttribute
from revsymg.graphs._adjancy_table import _AdjsT, _build_rev_sym_adj_table
from revsymg.graphs.revsym_edges import Edges
from revsymg.graphs.revsym_vertices import Vertices


# from revsymg.graphs.view.sub_edges import SubEdges
# from revsymg.graphs.view.sub_graph import SubRevSymGraph
# from revsymg.graphs.view.sub_vertices import SubVertices


# ============================================================================ #
#                                OVERLAPS GRAPH                                #
# ============================================================================ #
class RevSymGraph():
    """Reverse symmetric graph class.

    This graph is considering each orientation for all vertices and for
    all edges.
    It can contain several weakly connected components, which can contain
    only one strand or both merged.
    """

    _attributes: dict[str, Any]

    def __init__(self, attrs: Optional[dict[str, Any]] = None):
        """Initializer."""
        # Adjancies table ---
        adjs: _AdjsT = _build_rev_sym_adj_table()
        # Vertices container ---
        self._vertices: Vertices = Vertices(adjs)
        # Edges container ---
        self._edges: Edges = Edges(adjs)
        # Attributes container ---
        self._attributes: dict[str, Any] = attrs if attrs is not None else {}

    # -*- Graph methods -*-

    def attr(self, attrname: str) -> Any:
        """Return the value associated to attribute `attrname`.

        Parameters
        ----------
        attrname : str
            Attribute's name

        Returns
        -------
        Any
            Attribute value corresponding to the attribute name

        Raises
        ------
        NoGraphAttribute
            When there is no attribute key
        """
        try:
            return self._attributes[attrname]
        except KeyError as exc:
            raise NoGraphAttribute(attrname) from exc

    def attrs(self) -> dict[str, Any]:
        """Return the dictionnary of all attributes.

        Returns
        -------
        dict
            Dictionnary of attribute name as key and their value
        """
        return self._attributes

    def set_attr(self, attrname: str, attrvalue: Any):
        """Change the value of the attributes else create them.

        Parameters
        ----------
        attrname : str
            Attribute's name
        attrvalue : Any
            Attribute's value
        """
        self._attributes[attrname] = attrvalue

    def is_attr(self, attrname: str) -> bool:
        """Return True if there is an attribute called attrname.

        Parameters
        ----------
        attrname : str
            Attribute's name answered

        Returns
        -------
        bool
            If there is an attribute called attrname
        """
        return attrname in self._attributes

    # -*- Subgraph methods -*-

    # def subgraph_from_vertices(self,
    #                         vertices: Iterable[IndOrT]) -> SubRevSymGraph:
    #     """Return a sub overlaps graph from vertices.

    #     Parameters
    #     ----------
    #     vertices : iterable of IndOrT
    #         An iterable of vertices

    #     Returns
    #     -------
    #     SubRevSymGraph
    #         The overlaps graph masked by vertices
    #     """
    #     sub_vertices = SubVertices(self._vertices)
    #     # REFACTOR: sub-edges should not have (sub-)vertices in parameter
    #     sub_edges = SubEdges(self._vertices, self._edges)
    #     for v in vertices:
    #         sub_vertices.add_to_mask(v)
    #     for v in sub_vertices:
    #         for w, e_ind in self._edges.succs(v):
    #             if w in sub_vertices:
    #                 sub_edges.add_to_mask(((v, w), e_ind))
    #     # REFACTOR: sub-vertex should get (sub-)vertex attributes, same for edges
    #     return SubRevSymGraph(
    #         sub_vertices, sub_edges,
    #         self._vertices._attributes, self._edges._attributes,
    #         self._attributes,
    #     )

    # def subgraph_from_edges(self,
    #                         edges: Iterable[EIndOrIndT]) -> SubRevSymGraph:
    #     """Return a sub overlaps graph from edges.

    #     Parameters
    #     ----------
    #     edges : iterable of EIndOrIndT
    #         An iterable of edges

    #     Returns
    #     -------
    #     SubRevSymGraph
    #         The overlaps graph masked by edges
    #     """
    #     sub_vertices = SubVertices(self._vertices)
    #     sub_edges = SubEdges(self._vertices, self._edges)
    #     for edge in edges:
    #         sub_vertices.add_to_mask(edge[EVERTICES][0])
    #         sub_vertices.add_to_mask(edge[EVERTICES][1])
    #         sub_edges.add_to_mask(edge)
    #     return SubRevSymGraph(
    #         sub_vertices, sub_edges,
    #         self._vertices._attributes, self._edges._attributes, self._attributes,
    #     )

    # -*- Getter -*-

    def vertices(self) -> Vertices:
        """Return the vertices container.

        Returns
        -------
        Vertices
            Vertices container
        """
        return self._vertices

    def edges(self) -> Edges:
        """Return the edge container.

        Returns
        -------
        Edges
            Edges container
        """
        return self._edges



# -*- coding=utf-8 -*-

"""Split strands vertices container class module."""

from __future__ import annotations

from typing import Any, Iterable, Iterator

from bitarray import bitarray

from revsymg.exceptions import (
    NoVertexIndex,
    NoVerticesAttribute,
    WrongAttributeType,
    _NoAttribute,
    _NoKey,
)
from revsymg.graphs._adjancy_table import _AdjsIndT, _shift_indices_adj_table
from revsymg.graphs.attributes import AttributesContainer
from revsymg.lib.index_lib import (
    IND,
    OR,
    PRED_IND,
    SUCC_IND,
    IndexT,
    IndOrT,
    OrT,
)


# DOC: do doc for SplitStrandsVertices
# ============================================================================ #
#                     REVERSE SYMMETRIC VERTICES CONTAINER                     #
# ============================================================================ #
class SplitStrandsVertices():
    r"""Unique oriented vertices container.

    This vertices container is containing only one orientation for each of its
    vertex *i.e.* :math:`\forall v \in V, \bar{v} \notin V`

    Warnings
    --------
    You must not call this class.
    Use it only for typing or get an already existing object.
    """

    def __init__(self, _adjs: _AdjsIndT, _cc: bitarray):
        """Initializer."""
        # Adjancy tables ---
        self._adjs: _AdjsIndT = _adjs
        # Describe the vertices' orientation
        #   [v_ind]: FORWARD_INT | REVERSE_INT
        self._cc: bitarray = _cc
        # Attributes for vertices
        self._attributes: AttributesContainer = AttributesContainer()

    # -*- Attributes -*-

    def attr(self, vertex_index: IndexT, attrname: str) -> Any:
        """Return the value associated to `vertex` attribute `attrname`.

        Parameters
        ----------
        vertex_index : IndexT
            Vertex's index
        attrname : str
            Attribute name

        Returns
        -------
        Any
            Attribute value corresponding to the attribute name

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        NoVerticesAttribute
            If there is no attribute nammed like this for vertices
        """
        try:
            return self._attributes.get(vertex_index, attrname)
        except _NoAttribute as exc:
            raise NoVerticesAttribute(attrname) from exc
        except _NoKey as exc:
            raise NoVertexIndex(vertex_index) from exc

    def attrs(self, vertex_index: IndexT) -> Iterator[tuple[str, Any]]:
        """Iterate over all vertex's attributes.

        Parameters
        ----------
        vertex_index: IndexT
            Vertex's index

        Yields
        ------
        str
            Attribute's name
        Any
            Attribute's value

        Raises
        ------
        NoVertexIndex
            If there is no vertex index
        """
        try:
            yield from self._attributes.get_all(vertex_index)
        except _NoKey as _no_key:
            raise NoVertexIndex(vertex_index) from _no_key

    def new_attr(self, attrname: str, default: Any):
        """Add an attribute entry.

        Parameters
        ----------
        attrname : str
            Attribute's name
        default : Any
            Default attribute value

        Note
        ----
        Before setting attribute to one vertex, it is necessary to add
        an attribute entry using this method.
        """
        self._attributes.new_attr(attrname, default)

    def set_attr(self, vertex_index: IndexT, attrname: str, attrvalue: Any):
        """Change the value of the attributes else create them.

        Parameters
        ----------
        vertex_index : IndexT
            Vertex's index
        attrname : str
            Attribute's name
        attrvalue : Any
            Attribute's value

        Raises
        ------
        NoVerticesAttribute
            If there is no attribute nammed like this for vertices
        WrongAttributeType
            Given value does not correspond to attribute's type
        NoVertexIndex
            If there is no vertex index

        Note
        ----
        Be sure having added an attribute entry using the method
        :meth:`~revsymg.Vertices.new_attr`
        """
        try:
            self._attributes.set_attr(vertex_index, attrname, attrvalue)
        except _NoAttribute as _no_attr:
            raise NoVerticesAttribute(attrname) from _no_attr
        except WrongAttributeType as wrong_attr_type:
            raise wrong_attr_type
        except _NoKey as _no_key:
            raise NoVertexIndex(vertex_index) from _no_key

    # -*- Orientation -*-

    def contains_indor(self, vertex: IndOrT) -> bool:
        """Answer if oriented vertex in vertices container.

        Parameters
        ----------
        vertex : IndOrT
            Oriented vertex

        Returns
        -------
        bool
            Oriented vertex in vertices container
        """
        if vertex[IND] < len(self._cc):
            return vertex[OR] == self._cc[vertex[IND]]
        return False

    def orientation(self, vertex_index: IndexT) -> OrT:
        """Return the orientation of vertex's index.

        Parameters
        ----------
        vertex_index : IndexT
            Vertex's index

        Returns
        -------
        OrT
            Vertex's index orientation

        Raises
        ------
        NoVertexIndex
            If there is no vertex index
        """
        try:
            return self._cc[vertex_index]  # type: ignore
        except IndexError as index_err:
            raise NoVertexIndex(vertex_index) from index_err

    # -*- Getter -*-

    def cc_record(self) -> bitarray:
        """Getter for connected component record.

        Returns
        -------
        bitarray
            For each index store the orientation which is in the arbitrarly
            choosen connected component

        Warnings
        --------
        You should not use this method, use instead:
        ```python
        >>> vertex in vertices
        ```

        :meta private:
        """
        # XXX: necessary?
        return self._cc

    # -*- Setter -*-

    def add(self, orientation: OrT) -> IndexT:
        """Add the vertex to the connected component record.

        Parameters
        ----------
        orientation : OrT
            New orientation to add to vertices container

        Returns
        -------
        IndexT
            Last vertex's index
        """
        self._cc.append(orientation)
        self._adjs[PRED_IND].append([])
        self._adjs[SUCC_IND].append([])
        return len(self._cc) - 1

    def delete(self, vertex_index: IndexT):
        """Remove vertex from vertices container.

        Parameters
        ----------
        vertex_index : IndexT
            Vertex to delete

        Raises
        ------
        NoVertexIndex
            If vertex does not exist

        Warning
        -------
        This operation may invalidate vertex indices.
        In fact, deleting one vertex implies shifting all next indices.
        Because of this,
        keys should always be removed in decreasing index order:

        >>> for vertex in sorted(vertices_to_del, reverse=True):
        >>>     graph.vertices().delete(vertex)

        Note
        ----
        All next vertex indices are decreased in edges they belong.
        But concerning edge indices are not.
        """
        #
        # Exceptions
        #
        if vertex_index >= len(self._cc):
            raise NoVertexIndex(vertex_index)
        #
        # Forward and reverse deletion
        #
        for index in range(vertex_index):
            for adj_ind in (PRED_IND, SUCC_IND):
                self._adjs[adj_ind][index] = list(
                    _shift_indices_adj_table(
                        self._adjs[adj_ind][index],
                        vertex_index,
                    ),
                )
        for index in range(vertex_index, len(self._cc) - 1):
            for adj_ind in (PRED_IND, SUCC_IND):
                self._adjs[adj_ind][index] = list(
                    _shift_indices_adj_table(
                        self._adjs[adj_ind][index + 1],
                        vertex_index,
                    ),
                )
        for adj_ind in (PRED_IND, SUCC_IND):
            self._adjs[adj_ind].pop()
        self._cc = self._cc[:vertex_index] + self._cc[vertex_index + 1:]
        self._attributes.delete_key(vertex_index)

    def delete_several(self, vertex_indices: Iterable[IndexT]):
        """Remove vertices from vertices container.

        Parameters
        ----------
        vertex_indices : iterable of IndexT
            Vertices to delete

        Raises
        ------
        NoVertexIndex
            If one vertex does not exist

        Warning
        -------
        This operation may invalidate vertex indices.
        In fact, deleting one vertex implies shifting all next indices.

        Notes
        -----
        At the oposite of :meth:`~revsymg.SplitStrandsVertices.delete`
        method, it is not necessary to sort `vertex_indices`.
        This is done internally.

        All next vertex indices are decreased in edges they belong.
        But concerning edge indices are not.
        """
        for vertex_index in sorted(vertex_indices, reverse=True):
            #
            # Exceptions
            #
            if vertex_index >= len(self._cc):
                raise NoVertexIndex(vertex_index)
            #
            # Forward and reverse deletion
            #
            for index in range(vertex_index):
                for adj_ind in (PRED_IND, SUCC_IND):
                    self._adjs[adj_ind][index] = list(
                        _shift_indices_adj_table(
                            self._adjs[adj_ind][index],
                            vertex_index,
                        ),
                    )
            for index in range(vertex_index, len(self._cc) - 1):
                for adj_ind in (PRED_IND, SUCC_IND):
                    self._adjs[adj_ind][index] = list(
                        _shift_indices_adj_table(
                            self._adjs[adj_ind][index + 1],
                            vertex_index,
                        ),
                    )
            for adj_ind in (PRED_IND, SUCC_IND):
                self._adjs[adj_ind].pop()
            self._cc = self._cc[:vertex_index] + self._cc[vertex_index + 1:]
            self._attributes.delete_key(vertex_index)

    # -*- Overloaded built-in -*-

    def __iter__(self) -> Iterator[IndexT]:
        """Iterate on oriented vertices.

        Yields
        ------
        IndexT
            Vertices
        """
        yield from range(len(self._cc))

    def __contains__(self, vertex: IndexT) -> bool:
        """Answer if vertex is in vertices container.

        Returns
        -------
        bool
            Vertex is in vertices container
        """
        return vertex < len(self._cc)

    def __len__(self) -> int:
        """Return the number of oriented vertices.

        Returns
        -------
        int
            Number of orient
        """
        return len(self._cc)



# -*- coding=utf-8 -*-

"""Filters utility module."""

from collections.abc import Callable, Hashable
from typing import Any


FilterT = Callable[..., bool]


def always_true(*args: tuple[Any, ...],  # pylint: disable=unused-argument
                **kwargs: dict[Hashable, Any]) -> bool:
    """Return always True value.

    Parameters
    ----------
    args : Any
        Anything
    kwargs : Any
        Anything

    Returns
    -------
    bool
        True
    """
    return True



# -*- coding=utf-8 -*-

"""Exceptions classes module."""

from __future__ import annotations

from typing import Any, Optional

from revsymg.lib.index_lib import IndexT, IndOrT
from revsymg.lib.str_lib import IdT


# ============================================================================ #
#                             ATTRIBUTES EXCEPTIONS                            #
# ============================================================================ #
class _NoKey(Exception):
    """NoKey class."""

    def __init__(self, key: IndexT):
        super().__init__()
        self.__key = key

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return f'There is no key `{self.__key}` in attributes container.'


# ---------------------------------------------------------------------------- #
#                              Vertices And Edges                              #
# ---------------------------------------------------------------------------- #
class _NoAttribute(Exception):
    """NoAttribute class."""

    _object_name: str = 'objects'

    def __init__(self, attrname: str):
        super().__init__()
        self.__attrname = attrname

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return (
            f'There is no attribute nammed `{self.__attrname}` for the'
            f' {self._object_name}.'
        )


class NoGraphAttribute(_NoAttribute):
    """NoGraphAttribute class for graph object."""

    _object_name: str = 'graph'


class NoVerticesAttribute(_NoAttribute):
    """NoVerticesAttribute class for vertices container."""

    _object_name: str = 'vertices'


class NoEdgesAttribute(_NoAttribute):
    """NoEdgesAttribute class for edges container."""

    _object_name: str = 'edges'


# ---------------------------------------------------------------------------- #
#                             Vertex And Edge Keys                             #
# ---------------------------------------------------------------------------- #
class WrongAttributeType(Exception):
    """WrongAttributeType class."""

    def __init__(self, attrvalue: Any, attrtype: type):
        """The Initializer."""
        super().__init__()
        self.__attrvalue = attrvalue
        self.__attrtype = attrtype

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return (
            f'`{self.__attrvalue}` is type `{type(self.__attrvalue).__name__}`'
            f' while type `{self.__attrtype.__name__}` is expected.'
        )


# ============================================================================ #
#                                   VERTICES                                   #
# ============================================================================ #
class NoVertexIndex(Exception):
    """Vertex does not exist."""

    def __init__(self, vertex_index: IndexT):
        """The Initializer."""
        super().__init__()
        self.__vertex_index = vertex_index

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return f'{self.__vertex_index} is not in the vertices container.'


# ============================================================================ #
#                                     EDGES                                    #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                  Edges Index                                 #
# ---------------------------------------------------------------------------- #
class NoEdgeIndex(Exception):
    """Exception class for NoEdgeIndex error."""

    def __init__(self, edge_index: IndexT):
        """The Initializer."""
        super().__init__()
        self.__edge_index = edge_index

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return f'`{self.__edge_index}` is not in the edge index.'


# ---------------------------------------------------------------------------- #
#                                  RevSymEdges                                 #
# ---------------------------------------------------------------------------- #
class NoEdge(Exception):
    """Exception class for NoEdge error."""

    def __init__(self, first_vertex: IndOrT, second_vertex: IndOrT,
                 edge_index: IndexT):
        """The Initializer."""
        super().__init__()
        self.__edge = (first_vertex, second_vertex)
        self.__edge_index = edge_index

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return (
            f'There is no edge `{self.__edge}` with edge index'
            f'`{self.__edge_index}` in the edges container.'
        )


# ---------------------------------------------------------------------------- #
#                                 IndicesEdges                                 #
# ---------------------------------------------------------------------------- #
class NoIndicesEdge(Exception):
    """Exception class for NoIndicesEdge error."""

    def __init__(self, first_vertex: IndexT, second_vertex: IndexT,
                 edge_index: IndexT):
        """The Initializer."""
        super().__init__()
        self.__edge = (first_vertex, second_vertex)
        self.__edge_index = edge_index

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return (
            f'There is no edge `{self.__edge}` with edge index'
            f'`{self.__edge_index}` in the edges container.'
        )


# ============================================================================ #
#                               READS IDENTIFIERS                              #
# ============================================================================ #
class NoReadId(Exception):
    """No read's identifier exception class."""

    def __init__(self, read_id: IdT):
        """The Initializer."""
        super().__init__()
        self.__read_id = read_id

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return f"There is no read's identifier `{self.__read_id}` in container."


# ============================================================================ #
#                             CONNECTED COMPONENTS                             #
# ============================================================================ #
class NoConnectedComponentId(Exception):
    """Exception for vertex with no connected component identifier."""

    def __init__(self, vertex: IndOrT):
        """The Initializer."""
        super().__init__()
        self.__vertex = vertex

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return (
            f'Vertex `{self.__vertex}` is not associated'
            ' with any connected component.'
        )


class NotNoneCCompId(Exception):
    """The NotNoneCCompId class."""

    def __init__(self, v: IndOrT, v_cc: Optional[int],
                 v_rev: IndOrT, v_rev_cc: Optional[int]):
        """The Initializer."""
        super().__init__()
        self.__v = v
        self.__v_cc = v_cc
        self.__v_rev = v_rev
        self.__v_rev_cc = v_rev_cc

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return (
            f'At least one of `{self.__v}` or `{self.__v_rev}` has no connected'
            f' component identifier (= None), see:\n'
            f'\t- cc id of `{self.__v}` = `{self.__v_cc}`\n'
            f'\t- cc id of `{self.__v_rev}` = `{self.__v_rev_cc}`'
        )



# -*- coding=utf-8 -*-

"""Global vertices container class module."""

from __future__ import annotations

from collections.abc import Iterator
from typing import Any, Iterable

from revsymg.exceptions import (
    NoVertexIndex,
    NoVerticesAttribute,
    WrongAttributeType,
    _NoAttribute,
    _NoKey,
)
from revsymg.graphs._adjancy_table import _AdjsT, _shift_rev_sym_adj_table
from revsymg.graphs.attributes import AttributesContainer
from revsymg.lib.index_lib import (
    FORWARD_INT,
    PRED_IND,
    REVERSE_INT,
    SUCC_IND,
    IndexT,
    IndOrT,
)


# ============================================================================ #
#                              VERTICES CONTAINER                              #
# ============================================================================ #
class Vertices():
    r"""Vertices container class for both vertices' orientations.

    For each raw read :math:`r \in \mathcal{R_{aw}}` there is one vertex for
    each orientation :math:`v \in V` and :math:`\bar{v} \in V`.

    Warnings
    --------
    You must not instantiate a Vertices object.
    Use it only for typing or get an already instanciated object.
    """

    # DOC: change read by something else

    def __init__(self, _adjs: _AdjsT):
        """The Initializer."""
        # Adjancy tables ---
        self._adjs: _AdjsT = _adjs
        # Attributes for vertices
        self._attributes: AttributesContainer = AttributesContainer()
        # Number of forward vertices
        self._card_vf: int = 0

    # -*- Attributes -*-

    def attr(self, vertex_index: IndexT, attrname: str) -> Any:
        """Return the value associated to `vertex` attribute `attrname`.

        Parameters
        ----------
        vertex_index : IndexT
            Vertex's index
        attrname : str
            Attribute name

        Returns
        -------
        Any
            Attribute value corresponding to the attribute name

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        NoVerticesAttribute
            If there is no attribute nammed like this for vertices
        """
        try:
            return self._attributes.get(vertex_index, attrname)
        except _NoAttribute as exc:
            raise NoVerticesAttribute(attrname) from exc
        except _NoKey as exc:
            raise NoVertexIndex(vertex_index) from exc

    def attrs(self, vertex_index: IndexT) -> Iterator[tuple[str, Any]]:
        """Iterate over all vertex's attributes.

        Parameters
        ----------
        vertex_index: IndexT
            Vertex's index

        Yields
        ------
        str
            Attribute's name
        Any
            Attribute's value

        Raises
        ------
        NoVertexIndex
            If there is no vertex index
        """
        try:
            yield from self._attributes.get_all(vertex_index)
        except _NoKey as _no_key:
            raise NoVertexIndex(vertex_index) from _no_key

    def new_attr(self, attrname: str, default: Any):
        """Add an attribute entry.

        Parameters
        ----------
        attrname : str
            Attribute's name
        default : Any
            Default attribute value

        Note
        ----
        Before setting attribute to one vertex, it is necessary to add
        an attribute entry using this method.
        """
        self._attributes.new_attr(attrname, default)

    def set_attr(self, vertex_index: IndexT, attrname: str, attrvalue: Any):
        """Change the value of the attributes else create them.

        Parameters
        ----------
        vertex_index : IndexT
            Vertex's index
        attrname : str
            Attribute's name
        attrvalue : Any
            Attribute's value

        Raises
        ------
        NoVerticesAttribute
            If there is no attribute nammed like this for vertices
        WrongAttributeType
            Given value does not correspond to attribute's type
        NoVertexIndex
            If there is no vertex index

        Note
        ----
        Be sure having added an attribute entry using the method
        :meth:`~revsymg.Vertices.new_attr`
        """
        try:
            self._attributes.set_attr(vertex_index, attrname, attrvalue)
        except _NoAttribute as _no_attr:
            raise NoVerticesAttribute(attrname) from _no_attr
        except WrongAttributeType as wrong_attr_type:
            raise wrong_attr_type
        except _NoKey as _no_key:
            raise NoVertexIndex(vertex_index) from _no_key

    # -*- Getter -*-

    def card_index(self) -> int:
        """Returns the number of distinct indices.

        Note
        ----

        The cardinality of indices set :math:`V_{ind}`
        is equal to :math:`|V_{ind}|`

        Returns
        -------
        int
            Number of distinct indices
        """
        return self._card_vf

    # -*- Setter -*-

    def add(self, number: int = 1) -> IndexT:
        """Add vertices to vertices container.

        Parameters
        ----------
        number : int, default 1
            Number of vertices to add

        Returns
        -------
        IndexT
            Last vertex's index (unoriented)
        """
        self._adjs[PRED_IND].extend([[] for _ in range(number)])
        self._adjs[SUCC_IND].extend([[] for _ in range(number)])
        self._attributes.add_keys(number)
        self._card_vf += number
        return self._card_vf - 1

    def delete(self, vertex_index: IndexT):
        """Remove vertex from vertices container.

        Parameters
        ----------
        vertex_index : IndexT
            Vertex to delete

        Raises
        ------
        NoVertexIndex
            If vertex does not exist

        Warning
        -------
        This operation may invalidate vertex indices.
        In fact, deleting one vertex implies shifting all next indices.
        Because of this,
        keys should always be removed in decreasing index order:

        >>> for vertex in sorted(vertices_to_del, reverse=True):
        >>>     graph.vertices().delete(vertex)

        Note
        ----
        All next vertex indices are decreased in edges they belong.
        But concerned edges' indices are not.
        """
        #
        # Exceptions
        #
        if vertex_index >= self._card_vf:
            raise NoVertexIndex(vertex_index)
        # OPTIMIZE: just move the last one to index, update its index and pop
        #
        # Forward and reverse deletion
        #
        for index in range(vertex_index):
            for adj_ind in (PRED_IND, SUCC_IND):
                self._adjs[adj_ind][index] = list(
                    _shift_rev_sym_adj_table(
                        self._adjs[adj_ind][index],
                        vertex_index,
                    ),
                )
        #
        # Update next vertices indices
        #
        for index in range(vertex_index, self._card_vf - 1):
            for adj_ind in (PRED_IND, SUCC_IND):
                self._adjs[adj_ind][index] = list(
                    _shift_rev_sym_adj_table(
                        self._adjs[adj_ind][index + 1],
                        vertex_index,
                    ),
                )
        for adj_ind in (PRED_IND, SUCC_IND):
            self._adjs[adj_ind].pop()
        self._card_vf -= 1
        self._attributes.delete_key(vertex_index)

    def delete_several(self, vertex_indices: Iterable[IndexT]):
        """Remove vertices from vertices container.

        Parameters
        ----------
        vertex_indices : iterable of IndexT
            Vertices to delete

        Raises
        ------
        NoVertexIndex
            If one vertex does not exist

        Warning
        -------
        This operation may invalidate vertex indices.
        In fact, deleting one vertex implies shifting all next indices.

        Notes
        -----
        At the oposite of :meth:`~revsymg.Vertices.delete`
        method, it is not necessary to sort `vertex_indices`.
        This is done internally.

        All next vertex indices are decreased in edges they belong.
        But concerned edges' indices are not.
        """
        for vertex_index in sorted(vertex_indices, reverse=True):
            #
            # Exceptions
            #
            if vertex_index >= self._card_vf:
                raise NoVertexIndex(vertex_index)
            #
            # Forward and reverse deletion
            #
            for index in range(vertex_index):
                for adj_ind in (PRED_IND, SUCC_IND):
                    self._adjs[adj_ind][index] = list(
                        _shift_rev_sym_adj_table(
                            self._adjs[adj_ind][index],
                            vertex_index,
                        ),
                    )
            for index in range(vertex_index, self._card_vf - 1):
                for adj_ind in (PRED_IND, SUCC_IND):
                    self._adjs[adj_ind][index] = list(
                        _shift_rev_sym_adj_table(
                            self._adjs[adj_ind][index + 1],
                            vertex_index,
                        ),
                    )
            for adj_ind in (PRED_IND, SUCC_IND):
                self._adjs[adj_ind].pop()
            self._card_vf -= 1
            self._attributes.delete_key(vertex_index)

    # -*- Overloaded built-in -*-

    def __iter__(self) -> Iterator[IndOrT]:
        """Iterate on oriented vertices.

        Yields
        ------
        IndOrT
            Oriented vertices
        """
        for ind in range(self._card_vf):
            yield ind, FORWARD_INT
            yield ind, REVERSE_INT

    def __contains__(self, vertex_index: IndexT) -> bool:
        """Return True if vertex index is in container, else False.

        Parameters
        ----------
        vertex_index : IndexT
            Vertex index

        Returns
        -------
        bool
            True if oriented vertex is in container, else False
        """
        return vertex_index < self._card_vf

    def __len__(self) -> int:
        """Return the number of oriented vertices.

        Returns
        -------
        int
            Number of oriented vertices
        """
        return 2 * self._card_vf



# -*- coding=utf-8 -*-

"""Protected adjancy symmetric reverse table module."""


from typing import Iterable

from revsymg.lib.index_lib import IND, OR, IndexT, IndIndT, IndOrIndT


# ============================================================================ #
#                                     TYPE                                     #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                            Reverse Symmetric Table                           #
# ---------------------------------------------------------------------------- #
_SubAdjT = list[IndOrIndT]

#   [PREDS_IND]: Predecessors lists
#       [v_index]: [(u_indor, e_ind) | (u_indor, v_indF) in E]
#   [SUCCS_IND]: Succesors lists
#       [u_index]: [(v_indor, e_ind) | (u_indF, v_indor) in E]
_AdjsT = tuple[
    list[_SubAdjT],
    list[_SubAdjT],
]


# ---------------------------------------------------------------------------- #
#                                 Indices Table                                #
# ---------------------------------------------------------------------------- #
_SubAdjIndT = list[IndIndT]

# [PREDS_IND]: Predecessors lists
#   [v_index]: [(u_ind, e_ind) | (u_ind, v_ind) in E]
# [SUCCS_IND]: Succesors lists
#   [u_index]: [(v_ind, e_ind) | (u_ind, v_ind) in E]
_AdjsIndT = tuple[
    list[_SubAdjIndT],
    list[_SubAdjIndT],
]


# ============================================================================ #
#                                   FUNCTION                                   #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                            Reverse Symmetric Table                           #
# ---------------------------------------------------------------------------- #
def _build_rev_sym_adj_table() -> _AdjsT:
    """Build empty symmetry reverse table for vertices/edges containers.

    Returns
    -------
    _AdjsT
        Symmetry reverse table
    """
    return ([], [])


def _shift_rev_sym_adj_table(sub_adj_tab: _SubAdjT,
                             pivot: IndexT) -> Iterable[IndOrIndT]:
    """Shift the sub adjancies table according pivot.

    Parameters
    ----------
    sub_adj_tab : _SubAdjT
        Sub-adjancies table
    pivot : IndexT
        Pivot index

    Yields
    ------
    IndOrIndT
        Shifted vertices indices
    """
    for indor, e_ind in sub_adj_tab:
        if indor[IND] != pivot:
            if indor[IND] > pivot:
                indor = (indor[IND] - 1, indor[OR])
            yield indor, e_ind


# ---------------------------------------------------------------------------- #
#                                 Indices Table                                #
# ---------------------------------------------------------------------------- #
def _build_indices_adj_table() -> _AdjsIndT:
    """Build empty indices table for vertices/edges containers.

    Returns
    -------
    _AdjsIndT
        Indices table
    """
    return ([], [])


def _shift_indices_adj_table(sub_adj_tab: _SubAdjIndT,
                             pivot: IndexT) -> Iterable[IndIndT]:
    """Shift the sub adjancies table according pivot.

    Parameters
    ----------
    sub_adj_tab : _SubAdjIndT
        Sub-adjancies table
    pivot : IndexT
        Pivot index

    Yields
    ------
    IndIndT
        Shifted vertices indices
    """
    for ind, e_ind in sub_adj_tab:
        if ind != pivot:
            if ind > pivot:
                ind = ind - 1
            yield ind, e_ind



# -*- coding=utf-8 -*-

"""Split strands overlaps graph class module."""

from __future__ import annotations

from typing import Any, Iterable, Optional

from bitarray import bitarray

from revsymg.exceptions import NoGraphAttribute
from revsymg.graphs._adjancy_table import _AdjsIndT, _build_indices_adj_table
from revsymg.graphs.split_strands_edges import SplitStrandsEdges
from revsymg.graphs.split_strands_vertices import SplitStrandsVertices
# from revsymg.graphs.view.sub_graph import SubRevSymGraph
from revsymg.lib.index_lib import EIndIndT, IndexT


# DOC: do doc for SplitStrandsGraph
# ============================================================================ #
#                        REVERSE SYMMETRIC OVERLAPS GRAPH                       #
# ============================================================================ #
class SplitStrandsGraph():
    r"""Graph containing only one DNA strand.

    For this graph :math:`G = (V, E)` we have:

    * :math:`\forall v \in V, \bar{v} \notin V`
    * :math:`\forall e \in E, \bar{e} \notin E`
    """

    def __init__(self, attrs: Optional[dict[str, Any]] = None):
        """Initializer."""
        # Adjancies table ---
        adjs: _AdjsIndT = _build_indices_adj_table()
        # Describe which orientation is in the arbitrary choosen connected comp.
        # - [v_ind]: FORWARD_INT | REVERSE_INT
        _cc: bitarray = bitarray()
        # Vertices container ---
        self._vertices: SplitStrandsVertices = SplitStrandsVertices(adjs, _cc)
        # Edges container ---
        self._edges: SplitStrandsEdges = SplitStrandsEdges(adjs, _cc)
        # Attributes container ---
        self._attributes: dict[str, Any] = attrs if attrs is not None else {}

    # -*- Graph methods -*-

    def attr(self, attrname: str) -> Any:
        """Return the value associated to attribute `attrname`.

        Parameters
        ----------
        attrname : str
            Attribute's name

        Returns
        -------
        Any
            Attribute value corresponding to the attribute name

        Raises
        ------
        NoGraphAttribute
            When there is no attribute key
        """
        try:
            return self._attributes[attrname]
        except KeyError as exc:
            raise NoGraphAttribute(attrname) from exc

    def attrs(self) -> dict[str, Any]:
        """Return the dictionnary of all attributes.

        Returns
        -------
        dict
            Dictionnary of attribute name as key and their value
        """
        return self._attributes

    def set_attr(self, attrname: str, attrvalue: Any):
        """Change the value of the attributes else create them.

        Parameters
        ----------
        attrname : str
            Attribute's name
        attrvalue : Any
            Attribute's value
        """
        self._attributes[attrname] = attrvalue

    # -*- Subgraph methods -*-

    # def subgraph_from_vertices(self, vertices: Iterable[IndexT]) -> SubRevSymGraph:
    #     """Return a sub overlaps graph from vertices.

    #     Parameters
    #     ----------
    #     vertices : iterable of IndexT
    #         An iterable of vertices

    #     Returns
    #     -------
    #     SubRevSymGraph
    #         The overlaps graph masked by vertices
    #     """
    #     raise NotImplementedError

    # def subgraph_from_edges(self,
    #                         edges: Iterable[EIndIndT]) -> SubRevSymGraph:
    #     """Return a sub overlaps graph from edges.

    #     Parameters
    #     ----------
    #     edges : iterable of EIndOrIndT
    #         An iterable of edges

    #     Returns
    #     -------
    #     SubRevSymGraph
    #         The overlaps graph masked by edges
    #     """
    #     raise NotImplementedError

    # -*- Getter -*-

    def vertices(self) -> SplitStrandsVertices:
        """Return the vertices container.

        Returns
        -------
        SplitStrandsVertices
            Vertices container
        """
        return self._vertices

    def edges(self) -> SplitStrandsEdges:
        """Return the edge container.

        Returns
        -------
        SplitStrandsEdges
            Edges container
        """
        return self._edges



# -*- coding=utf-8 -*-

"""Merged strands overlaps graph class module."""

from __future__ import annotations

from revsymg.graphs.revsym_graph import RevSymGraph


# DOC: do doc for MergedStrandsRevSymGraph
# ============================================================================ #
#                         MERGED STRANDS OVERLAPS GRAPH                        #
# ============================================================================ #
MergedStrandsRevSymGraph = RevSymGraph
r"""Weakly connected overlaps graph containing a merge of DNA strands.

This graph :math:`G = (V, E)` contains only one weakly connected component where
there are the two DNA strands.
Thus:

- :math:`\forall v \in V, \bar{v} \in V`
- :math:`\forall e \in E, \bar{e} \in E`
"""



# -*- coding=utf-8 -*-

"""Just for module."""

# flake8: noqa

from revsymg.algorithms.connected_components import *
from revsymg.algorithms.graph_functions import *
from revsymg.algorithms.transitive_reduction import *



# -*- coding=utf-8 -*-

"""Algorithms for Symmetric Overlaps Graph."""

from __future__ import annotations

from pathlib import Path
from typing import Union

from networkx import DiGraph as nx_DiGraph
from networkx import write_graphml as nx_write_graphml

from revsymg.exceptions import NoVertexIndex
from revsymg.graphs.merged_strands_graph import MergedStrandsRevSymGraph
from revsymg.graphs.revsym_graph import RevSymGraph
from revsymg.lib.index_lib import IND, OR, IndOrT
from revsymg.lib.str_lib import STRAND_STR


# ============================================================================ #
#                                    DEGREE                                    #
# ============================================================================ #
def in_degree(graph: Union[RevSymGraph, MergedStrandsRevSymGraph],
              vertex: IndOrT) -> int:
    """Return the in-degree of `vertex`.

    Parameters
    ----------
    graph : RevSymGraph or MergedStrandsRevSymGraph
        Reverse symmetric graph
    vertex : IndOrT
        Oriented vertex

    Returns
    -------
    int
        In-degree

    Raises
    ------
    NoVertexIndex
        If vertex does not exist in graph
    """
    try:
        return sum(1 for _ in graph.edges().preds(vertex))
    except NoVertexIndex as no_vertex_index:
        raise no_vertex_index


def out_degree(graph: Union[RevSymGraph, MergedStrandsRevSymGraph],
               vertex: IndOrT) -> int:
    """Return the out-degree of `vertex`.

    Parameters
    ----------
    graph : RevSymGraph or MergedStrandsRevSymGraph
        Reverse symmetric graph
    vertex : IndOrT
        Oriented vertex

    Returns
    -------
    int
        Out-degree

    Raises
    ------
    NoVertexIndex
        If vertex does not exist in graph
    """
    try:
        return sum(1 for _ in graph.edges().succs(vertex))
    except NoVertexIndex as no_vertex_index:
        raise no_vertex_index


def degree(graph: Union[RevSymGraph, MergedStrandsRevSymGraph],
           vertex: IndOrT) -> int:
    """Return the degree of `vertex`.

    Parameters
    ----------
    graph : RevSymGraph or MergedStrandsRevSymGraph
        Reverse symmetric graph
    vertex : IndOrT
        Oriented vertex

    Returns
    -------
    int
        Degree

    Raises
    ------
    NoVertexIndex
        If vertex does not exist in graph
    """
    try:
        return in_degree(graph, vertex) + out_degree(graph, vertex)
    except NoVertexIndex as no_vertex_index:
        raise no_vertex_index


# ============================================================================ #
#                           INPUTS-OUTPUTS FUNCTIONS                           #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                               Outputs Function                               #
# ---------------------------------------------------------------------------- #
def revsymg_to_graphml(graph: RevSymGraph, path: Path):
    """Save the graph as graphml format.

    Parameters
    ----------
    graph : RevSymGraph
        Reverse symmetric graph to output in graphml file format
    path : Path
        Output graphml path
    """
    vertices = graph.vertices()
    edges = graph.edges()
    di_graph = nx_DiGraph()
    for vertex in vertices:
        vertex_id = f'{vertex[IND]}{STRAND_STR[vertex[OR]]}'
        attrs = dict(vertices.attrs(vertex[IND]))
        di_graph.add_node(
            vertex_id, **attrs,
        )
    for u, v, e_ind in edges:
        u_id = f'{u[IND]}{STRAND_STR[u[OR]]}'
        v_id = f'{v[IND]}{STRAND_STR[v[OR]]}'
        attrs = dict(edges.attrs(e_ind))
        di_graph.add_edge(
            u_id, v_id, **attrs,
        )
    # Write graphml
    nx_write_graphml(di_graph, path)



# -*- coding=utf-8 -*-

"""Library for overlaps graph objects."""

from __future__ import annotations

from typing import Union

from revsymg.graphs.merged_strands_graph import MergedStrandsRevSymGraph
from revsymg.graphs.revsym_graph import RevSymGraph
from revsymg.graphs.split_strands_graph import SplitStrandsGraph


# from revsymg.graphs.view.sub_graph import SubRevSymGraph


# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
GraphsT = Union[
    RevSymGraph,
    # SubRevSymGraph,
    MergedStrandsRevSymGraph,
    SplitStrandsGraph,
]
GraphsCnntedCompT = Union[
    SplitStrandsGraph,
    MergedStrandsRevSymGraph,
]



# -*- coding=utf-8 -*-

"""Split strands edges container class module."""

from __future__ import annotations

from typing import Any, Iterator

from bitarray import bitarray

from revsymg.exceptions import (
    NoEdgeIndex,
    NoEdgesAttribute,
    NoIndicesEdge,
    NoVertexIndex,
    WrongAttributeType,
    _NoAttribute,
    _NoKey,
)
from revsymg.graphs._adjancy_table import _AdjsIndT
from revsymg.graphs.attributes import AttributesContainer
from revsymg.lib.index_lib import (
    PRED_IND,
    SUCC_IND,
    EIndIndT,
    EIndT,
    IndexT,
    IndIndT,
)


# DOC: do doc for SplitStrandsEdges
# ============================================================================ #
#                         SPLIT STRANDS EDGES CONTAINER                        #
# ============================================================================ #
class SplitStrandsEdges():
    r"""Unique oriented edges container.

    This edges container is containing only one orientation for each of its
    edge *i.e.* :math:`\forall e \in E, \bar{e} \notin E`

    Warnings
    --------
    You must not instantiate a Edges object.
    Use it only for typing or get an already instanciated object.
    """

    def __init__(self, _adjs: _AdjsIndT, _cc: bitarray):
        """The Initializer."""
        # The number of current edges
        self._card_e: int = 0
        # The last edges integer identifier ---
        self._ind_e: IndexT = -1
        # Adjancies table ---------------------------------
        #
        # [PREDS_IND]: Predecessors lists
        #   [v_index]: [(u_ind, e_ind) | (u_ind, v_ind) in E]
        # [SUCCS_IND]: Succesors lists
        #   [u_index]: [(v_ind, e_ind) | (u_ind, v_ind) in E]
        #
        self._adjs: _AdjsIndT = _adjs
        # Describe the vertices' orientation
        #   [v_ind]: FORWARD_INT | REVERSE_INT
        self._cc: bitarray = _cc  # XXX: just for cheking exc but not necessary
        # Attributes for edges
        self._attributes: AttributesContainer = AttributesContainer()

    # -*- Attributes -*-

    def attr(self, edge_index: IndexT, attrname: str) -> Any:
        """Return the edge attribute `attrname`.

        Parameters
        ----------
        edge_index: IndexT
            Edge's index
        attrname: str
            Attribute name

        Returns
        -------
        Any
            Attribute value

        Raises
        ------
        NoEdgeIndex
            If edge does not exist
        NoEdgesAttribute
            If there is no attribute name like this for edges

        Note
        ----
        A removed edge still have a recorded attribute.
        """
        try:
            return self._attributes.get(edge_index, attrname)
        except _NoAttribute as exc:
            raise NoEdgesAttribute(attrname) from exc
        except _NoKey as exc:
            raise NoEdgeIndex(edge_index) from exc

    def attrs(self, edge_index: IndexT) -> Iterator[tuple[str, Any]]:
        """Iterate over all edge's attributes.

        Parameters
        ----------
        edge_index: IndexT
            Edge's index

        Yields
        ------
        str
            Attribute's name
        Any
            Attribute's value

        Raises
        ------
        NoEdgeIndex
            If there is no edge index

        Note
        ----
        A removed edge still have a recorded attribute.
        """
        try:
            yield from self._attributes.get_all(edge_index)
        except _NoKey as _no_key:
            raise NoEdgeIndex(edge_index) from _no_key

    def new_attr(self, attrname: str, default: Any):
        """Add an attribute entry.

        Parameters
        ----------
        attrname : str
            Attribute's name
        default : Any
            Default attribute value

        Note
        ----
        Before setting attribute to one edge, it is necessary to add
        an attribute entry using this method.
        """
        self._attributes.new_attr(attrname, default)

    def set_attr(self, edge_index: IndexT, attrname: str, attrvalue: Any):
        """Change the value of one attribute.

        Parameters
        ----------
        edge_index : IndexT
            Edge's index
        attrname : str
            Attribute's name
        attrvalue : Any
            Attribute's value

        Raises
        ------
        NoEdgesAttribute
            If there is no attribute nammed like this for edges
        WrongAttributeType
            Given value does not correspond to attribute's type
        NoEdgeIndex
            If there is no vertex index

        Note
        ----
        Be sure having added an attribute entry using the method
        :meth:`~revsymg.Vertices.new_attr`
        """
        try:
            self._attributes.set_attr(edge_index, attrname, attrvalue)
        except _NoAttribute as _no_attr:
            raise NoEdgesAttribute(attrname) from _no_attr
        except WrongAttributeType as wrong_attr_type:
            raise wrong_attr_type
        except _NoKey as _no_key:
            raise NoEdgeIndex(edge_index) from _no_key

    # -*- Getter -*-

    def eindor_to_eind(self, first_vertex: IndexT,
                       second_vertex: IndexT) -> Iterator[IndexT]:
        """Iterate over edge index from indices edge.

        Parameters
        ----------
        first_vertex : IndexT
            First vertex
        second_vertex : IndexT
            Second vertex

        Yields
        ------
        IndexT
            Edge index

        Raises
        ------
        NoVertexIndex
            One of the vertex in edge does not exist
        """
        #
        # Exceptions
        #
        if first_vertex >= len(self._cc):
            raise NoVertexIndex(first_vertex)
        if second_vertex >= len(self._cc):
            raise NoVertexIndex(second_vertex)
        #
        # Search
        #
        yield from (
            e_ind
            for succ, e_ind in self._adjs[SUCC_IND][first_vertex]
            if succ == second_vertex
        )

    # -*- Edges -*-

    def preds(self, vertex: IndexT) -> Iterator[IndIndT]:
        """Iterate over vertex's predecessors.

        Parameters
        ----------
        vertex: IndexT
            Vertex's index

        Yields
        ------
        IndIndT
            Predecessor vertex's index with edge index

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        """
        try:
            preds = self._adjs[PRED_IND][vertex]
        except IndexError as exc:
            raise NoVertexIndex(vertex) from exc
        yield from preds

    def succs(self, vertex: IndexT) -> Iterator[IndIndT]:
        """Iterate over vertex's successors.

        Parameters
        ----------
        vertex: IndexT
            Vertex's index

        Yields
        ------
        IndIndT
            Successors vertex's index with edge index

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        """
        try:
            succs = self._adjs[SUCC_IND][vertex]
        except IndexError as exc:
            raise NoVertexIndex(vertex) from exc
        yield from succs

    def neighbours(self, vertex: IndexT) -> Iterator[IndIndT]:
        """Iterate over `vertex`'s adjacencies vertices.

        Parameters
        ----------
        vertex: IndexT
            Vertex's index

        Yields
        ------
        IndIndT
            Adjacent vertices's index with edge index

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        """
        try:
            yield from self.preds(vertex)
            yield from self.succs(vertex)
        except NoVertexIndex as no_vertex_index:
            raise no_vertex_index

    # -*- Setter -*-

    def add(self, first_vertex: IndexT, second_vertex: IndexT) -> IndexT:
        """Add `edge` to edges.

        Parameters
        ----------
        first_vertex : IndexT
            First vertex
        second_vertex : IndexT
            Second vertex

        Returns
        -------
        IndexT
            Edge index

        Raises
        ------
        NoVertexIndex
            If one vertex does not exist
        """
        #
        # Exceptions
        #
        if len(self._adjs[PRED_IND]) <= first_vertex:
            raise NoVertexIndex(first_vertex)
        if len(self._adjs[PRED_IND]) <= second_vertex:
            raise NoVertexIndex(second_vertex)
        self._ind_e += 1
        #
        # Successor
        #
        self._adjs[SUCC_IND][first_vertex].append(
            (second_vertex, self._ind_e),
        )
        #
        # Predecessor
        #
        self._adjs[PRED_IND][second_vertex].append(
            (first_vertex, self._ind_e),
        )
        self._card_e += 1
        self._attributes.add_keys(1)
        return self._ind_e

    def delete(self, first_vertex: IndexT, second_vertex: IndexT,
               edge_index: IndexT):
        r"""Remove the edge.

        Parameters
        ----------
        first_vertex : IndexT
            First vertex
        second_vertex : IndexT
            Second vertex
        edge_index : IndexT
            Edge's index

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        NoIndicesEdge
            If there is no recorded edge with these vertices and this index

        Note
        ----
        When deleting, the edge is replacing by the last one in adjancy list,
        and the last element of the adjancy list is removed.
        """
        #
        # Exceptions
        #
        if first_vertex >= len(self._adjs[PRED_IND]):
            raise NoVertexIndex(first_vertex)
        if second_vertex >= len(self._adjs[PRED_IND]):
            raise NoVertexIndex(second_vertex)
        #
        # Delete u from v's predecessors
        #
        u_adj: IndIndT = (first_vertex, edge_index)
        try:
            adj_ind = self._adjs[PRED_IND][second_vertex].index(u_adj)
            # by symmetry, other index method bellow will raises ValueError too
        except ValueError as value_error:
            raise NoIndicesEdge(
                first_vertex, second_vertex, edge_index,
            ) from value_error
        if adj_ind < len(self._adjs[PRED_IND][second_vertex]) - 1:
            self._adjs[PRED_IND][second_vertex][adj_ind] = (
                self._adjs[PRED_IND][second_vertex].pop()
            )
        else:
            self._adjs[PRED_IND][second_vertex].pop()
        #
        # Delete v from u's successors
        #
        v_adj: IndIndT = (second_vertex, edge_index)
        adj_ind = self._adjs[SUCC_IND][first_vertex].index(v_adj)
        if adj_ind < len(self._adjs[SUCC_IND][first_vertex]) - 1:
            self._adjs[SUCC_IND][first_vertex][adj_ind] = (
                self._adjs[SUCC_IND][first_vertex].pop()
            )
        else:
            self._adjs[SUCC_IND][first_vertex].pop()
        self._card_e -= 1

    # -*- Overloaded built-in -*-

    def __iter__(self) -> Iterator[EIndIndT]:
        """Iterate over indices edges with edge index.

        Yields
        ------
        EIndIndT
            Indices edges with their edge index
        """
        for v in range(len(self._cc)):
            for w, e_ind in self._adjs[SUCC_IND][v]:
                yield v, w, e_ind

    def __contains__(self, edge: EIndT) -> bool:
        """Return True if the indices edge is in edges container.

        Parameters
        ----------
        edge: EIndT
            Indices edge

        Returns
        -------
        bool
            Indices edge in edges container
        """
        u, v = edge
        #
        # No vertices
        #
        if len(self._cc) <= u:
            return False
        if len(self._cc) <= v:
            return False
        #
        # Search in forward's successors table
        #
        return any(v == succ for succ, _ in self._adjs[SUCC_IND][u])

    def __len__(self) -> int:
        r"""Return the number of edges.

        Returns
        -------
        int
            Number of edges

        Notes
        -----
        :math:`\forall e \in E, \bar{e} \notin E`
        """
        return self._card_e



# -*- coding=utf-8 -*-

"""Connected component identifiers containers classes module."""

from __future__ import annotations

from typing import Iterator, Optional

from revsymg.exceptions import NoConnectedComponentId, NoVertexIndex
from revsymg.lib.index_lib import (
    FORWARD_INT,
    FWDOR_REVOR,
    IND,
    OR,
    ORIENT_REV,
    REVERSE_INT,
    IndexT,
    IndOrT,
    OrT,
)


# ============================================================================ #
#                     REVERSE SYMMETRIC CONNECTED COMPONENT                    #
# ============================================================================ #
class _RevSymCCIdFactory():
    """Connected component identifiers container builder.

    Warnings
    --------
    You must not use this class.
    """

    def __init__(self, card_index: int):
        """The Initializer."""
        # Number of disctint connected components
        self._nb_distinct: int = 0
        # Current couple of connected component identifiers
        self._ccompids_cur: tuple[IndexT, IndexT] = (0, 1)
        # Index of current couple of connected component identifier
        self._ccompid_ind: IndexT = -1
        # List of couples of connected component identifiers
        self._cc_couples: list[tuple[IndexT, IndexT]] = []
        # List of forward vertices connected component identifier
        self._f_vertices_cc: list[Optional[tuple[IndexT, OrT]]] = [
            None for _ in range(card_index)
        ]

    def new_cc_couple(self):
        """Set the couple of connected component Ids."""
        self._ccompid_ind += 1
        self._ccompids_cur = (
            self._nb_distinct, self._nb_distinct + 1,
        )
        self._cc_couples.append(self._ccompids_cur)
        self._nb_distinct += 2

    def set_cc_id(self, vertex: IndOrT):
        """Set the connected component identifiers to `vertex` and its reverse.

        Parameters
        ----------
        vertex : IndOrT
            Oriented vertex

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        """
        try:
            self._f_vertices_cc[vertex[IND]] = (self._ccompid_ind, vertex[OR])
        except IndexError as index_err:
            raise NoVertexIndex(vertex[IND]) from index_err

    def continuity(self, vertex: IndOrT) -> bool:
        """Return True if the continuity is respected.

        Parameters
        ----------
        vertex : IndOrT
            Oriented vertex

        Returns
        -------
        bool
            True if both strands are not in the same connected component

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        NoConnectedComponentId
            If vertex is not associated with any connected component
        """
        try:
            cc_indor = self._f_vertices_cc[vertex[IND]]
        except IndexError as index_err:
            raise NoVertexIndex(vertex[IND]) from index_err
        if cc_indor is None:
            raise NoConnectedComponentId(vertex)
        return (
            self._cc_couples[cc_indor[IND]][
                FWDOR_REVOR[vertex[OR]][cc_indor[OR]]
            ]
            == self._ccompids_cur[FORWARD_INT]
        )

    def merge_cc(self):
        """In case the two strands are confused, the ccomp ids are the equal."""
        self._nb_distinct -= 1
        self._ccompids_cur = (
            self._ccompids_cur[FORWARD_INT],
            self._ccompids_cur[FORWARD_INT],
        )
        self._cc_couples[self._ccompid_ind] = self._ccompids_cur

    def to_readonly_view(self) -> RevSymCCId:
        """Return a read-only view.

        Returns
        -------
        RevSymCCId
            Connected component identifier container
        """
        return RevSymCCId(
            self._nb_distinct, self._cc_couples, self._f_vertices_cc,
        )

    def __getitem__(self, vertex: IndOrT) -> Optional[IndexT]:
        """Return the cc number of oriented vertex.

        Parameters
        ----------
        vertex : IndOrT
            oriented vertex

        Returns
        -------
        IndexT or None
            Connected component identifier if exist or None

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        """
        try:
            cc_indor = self._f_vertices_cc[vertex[IND]]
        except IndexError as index_err:
            raise NoVertexIndex(vertex[IND]) from index_err
        if cc_indor is not None:
            return self._cc_couples[cc_indor[IND]][
                FWDOR_REVOR[vertex[OR]][cc_indor[OR]]
            ]
        return None


class RevSymCCId():
    """Connected Component Identifiers Container."""

    def __init__(self, _nb_distinct: int,
                 _cc_couples: list[tuple[IndexT, IndexT]],
                 _f_vertices_cc: list[Optional[tuple[IndexT, OrT]]],
                 ):
        """The Initializer."""
        # Number of disctint connected components
        self._nb_distinct: int = _nb_distinct
        # List of couples of connected component identifiers
        self._cc_couples: list[tuple[IndexT, IndexT]] = _cc_couples
        # List of forward vertices connected component identifier
        self._f_vertices_cc: list[Optional[tuple[IndexT, OrT]]] = (
            _f_vertices_cc
        )

    def card_cc(self) -> int:
        """Return the number of distinct connected components.

        Returns
        -------
        int
            Number of distinct connected components.
        """
        return self._nb_distinct

    def merged_orientations(self, vertex_index: IndexT) -> bool:
        """Return True if reversed is in the same connected component.

        Parameters
        ----------
        vertex_index : IndOrT
            Vertex index

        Returns
        -------
        bool
            Return True if the two orientations of a vertex
            are in the same connected component

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        NoConnectedComponentId
            If vertex is not associated with any connected component
        """
        try:
            cc_indor = self._f_vertices_cc[vertex_index]
        except IndexError as index_err:
            raise NoVertexIndex(vertex_index) from index_err
        if cc_indor is None:
            raise NoConnectedComponentId((vertex_index, FORWARD_INT))
        return (
            self._cc_couples[cc_indor[IND]][FORWARD_INT]
            == self._cc_couples[cc_indor[IND]][REVERSE_INT]
        )

    def reverse_cc(self, vertex: IndOrT) -> Optional[IndexT]:
        """Return the cc number of the `vertex` reverse.

        Parameters
        ----------
        vertex : IndOrT
            Oriented vertex

        Returns
        -------
        IndexT or None
            The connected component number (if exists, else None)

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        """
        try:
            cc_indor = self._f_vertices_cc[vertex[IND]]
        except IndexError as index_err:
            raise NoVertexIndex(vertex[IND]) from index_err
        if cc_indor is not None:
            return self._cc_couples[cc_indor[IND]][
                FWDOR_REVOR[ORIENT_REV[vertex[OR]]][cc_indor[OR]]
            ]
        return None

    def __getitem__(self, vertex: IndOrT) -> Optional[IndexT]:
        """Return the cc number of oriented vertex.

        Parameters
        ----------
        vertex : IndOrT
            oriented vertex

        Returns
        -------
        IndexT or None
            Connected component identifier if exist or None

        Raises
        ------
        NoVertexIndex
            If vertex does not exist
        """
        try:
            cc_indor = self._f_vertices_cc[vertex[IND]]
        except IndexError as index_err:
            raise NoVertexIndex(vertex[IND]) from index_err
        if cc_indor is not None:
            return self._cc_couples[cc_indor[IND]][
                FWDOR_REVOR[vertex[OR]][cc_indor[OR]]
            ]
        return None

    def __iter__(self) -> Iterator[
        tuple[
            IndOrT, Optional[IndexT],
            IndOrT, Optional[IndexT],
        ],
    ]:
        """Iter on both forward reverse vertices and their cc identifier.

        Yields
        ------
        IndOrT, int or None, IndorT, int or None
            Forward and reverse vertices
            and their connected component identifier
        """
        for v_ind, cc_indor in enumerate(self._f_vertices_cc):
            if cc_indor is None:
                v_cc = None
                v_rev_cc = None
            else:
                v_cc = self._cc_couples[cc_indor[IND]][cc_indor[OR]]
                v_rev_cc = self._cc_couples[cc_indor[IND]][
                    ORIENT_REV[cc_indor[OR]]
                ]
            yield (
                (v_ind, FORWARD_INT), v_cc,
                (v_ind, REVERSE_INT), v_rev_cc,
            )



# -*- coding=utf-8 -*-

"""The connected component algorithms module."""

from __future__ import annotations

from queue import Queue
from typing import Optional

from revsymg.algorithms.connected_components.cnnted_comp_id import (
    RevSymCCId,
    _RevSymCCIdFactory,
)
from revsymg.exceptions import NotNoneCCompId
from revsymg.graphs.graphs_lib import GraphsCnntedCompT, GraphsT
from revsymg.graphs.merged_strands_graph import MergedStrandsRevSymGraph
from revsymg.graphs.revsym_graph import RevSymGraph
from revsymg.graphs.split_strands_graph import SplitStrandsGraph
from revsymg.graphs.split_strands_vertices import SplitStrandsVertices
# from revsymg.graphs.view.sub_graph import SubRevSymGraph
from revsymg.lib.index_lib import IND, OR, IndOrT


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                        Redundant Connected Components                        #
# ---------------------------------------------------------------------------- #
def connected_components(graph: RevSymGraph) -> RevSymCCId:
    """Return the cc of each vertex.

    Parameters
    ----------
    graph : RevSymGraph
        The reverse symmetric graph

    Returns
    -------
    RevSymCCId
        A connected component identifiers container
    """
    ccompcont = _RevSymCCIdFactory(graph.vertices().card_index())
    for u in graph.vertices():
        if ccompcont[u] is None:  # Not explored nor its reverse yet
            ccompcont.new_cc_couple()
            __explore_weak(graph, u, ccompcont)

    return ccompcont.to_readonly_view()


def __explore_weak(graph: RevSymGraph, u: IndOrT,
                   ccompcont: _RevSymCCIdFactory):
    """Explore the `cc`-th connected component from `u`."""
    edges = graph.edges()
    lifo: Queue = Queue()
    lifo.put(u)
    # Add num cc
    ccompcont.set_cc_id(u)
    while not lifo.empty():
        u = lifo.get()
        for v, _ in edges.neighbours(u):
            # Not explored nor its reverse yet
            if ccompcont[v] is None:
                # Add num cc
                ccompcont.set_cc_id(v)
                lifo.put(v)
            # Reverse of an already explored: two strands are confused in cc
            elif not ccompcont.continuity(v):
                ccompcont.merge_cc()


# ---------------------------------------------------------------------------- #
#                      Non-redundant Connected Components                      #
# ---------------------------------------------------------------------------- #
# def nonredundant_connected_components(
#         ccompcont: RevSymCCId) -> CnntedComplistT:
#     r"""Return the connected components without redundance.

#     Requires that :math:`\forall v \in V, cc(v) \neq None`

#     - Use only one oriented
#       :class:`~revsymg.algorithms.connected_components.cnnted_comp_cls.SplitStrandsCnntedComp`
#       between two existing *i.e.* if :math:`cc(v) != cc(v')` then only keep
#       one connected component (because the other can be found from the other)

#     - Use :class:`~revsymg.algorithms.connected_components.cnnted_comp_cls.MergedStrandsCnntedComp`
#       in which it is known that :math:`v` and :math:`v'` are in the same
#       connected component

#     Parameters
#     ----------
#     ccompcont : RevSymCCId
#         The connected component identifier container

#     Returns
#     -------
#     list of SplitStrandsCnntedComp or MergedStrandsCnntedComp
#         List of simple or double connected component

#     Raises
#     ------
#     NotNoneCCompId
#         If :math:`\exists v \in V \mid cc(v) = None`
#     """  # noqa: E501 (too long line)
#     #
#     # Initialisation
#     #
#     n_redundant_cc = ccompcont.card_cc()
#     cc_to_keep: list[Optional[bool]] = [None for _ in range(n_redundant_cc)]
#     n_new_cc = 0  # [num_cc original] = new num_cc
#     #
#     # l_cc_trans[cc_ind]:
#     #   = None if cc_ind not yet discovered
#     #   = new_cc_ind >= 0 if already cc_ind already discovered
#     #   = -1 if cc_ind is not kept
#     l_cc_trans: list[Optional[int]] = [None for _ in range(n_redundant_cc)]
#     l_ccs: CnntedComplistT = []  # connected comp. kept
#     #
#     # Loop on vertices and their cc id
#     #
#     for v, v_cc, v_rev, v_rev_cc in ccompcont:
#         #
#         # cc ids must be not None
#         #
#         if v_cc is None or v_rev_cc is None:
#             raise NotNoneCCompId(v, v_cc, v_rev, v_rev_cc)
#         #
#         # Undiscovered cc
#         #
#         if cc_to_keep[v_cc] is None:
#             cc_to_keep[v_cc] = True
#             l_cc_trans[v_cc] = n_new_cc
#             #
#             # Simple cc
#             #
#             if v_rev_cc != v_cc:
#                 l_ccs.append(SplitStrandsCnntedComp(n_new_cc, v))
#                 cc_to_keep[v_rev_cc] = False
#             #
#             # Double cc
#             #
#             else:
#                 l_ccs.append(MergedStrandsCnntedComp(n_new_cc, v))
#             n_new_cc += 1
#         #
#         # Already discovered cc
#         #
#         else:
#             if cc_to_keep[v_cc]:
#                 new_num_cc: int = l_cc_trans[v_cc]  # type: ignore
#                 l_ccs[new_num_cc].add(v)
#             elif cc_to_keep[v_rev_cc]:
#                 # This is elif because we don't want to add twice v in double cc
#                 #   but add v in the kept complementary simple cc
#                 new_num_cc: int = l_cc_trans[v_rev_cc]  # type: ignore
#                 l_ccs[new_num_cc].add(v_rev)
#     return l_ccs


# def connected_comp_to_graphs(graph: RevSymGraph,
#                              ccompcont: RevSymCCId,
#                              ) -> list[GraphsCnntedCompT]:
#     r"""Returns one graph for each connected component.

#     Requires that :math:`\forall v \in V, cc(v) \neq None`

#     Parameters
#     ----------
#     graph : RevSymGraph
#         Reverse symmetric graph
#     ccompcont : RevSymCCId
#         Connected component identifiers container

#     Returns
#     -------
#     list of SplitStrandsGraph or MergedStrandsRevSymGraph
#         One graph for each (symmetric or merged) connected component

#     Raises
#     ------
#     NotNoneCCompId
#         If :math:`\exists v \in V \mid cc(v) = None`
#     """
#     # TODO: indices of vertices will change (and those of edges too)
#     #   Find a way to associate old and new indices
#     #
#     # Initialisation
#     #
#     n_redundant_cc = ccompcont.card_cc()
#     n_new_cc = 0  # [num_cc original] = new num_cc
#     #
#     # l_cc_trans[cc_ind]:
#     #   = None if cc_ind not yet discovered
#     #   = new_cc_ind >= 0 if already cc_ind already discovered
#     #   = -1 if cc_ind is not kept
#     l_cc_trans: list[Optional[int]] = [None for _ in range(n_redundant_cc)]
#     graph_ccs: list[GraphsCnntedCompT] = []
#     #
#     # Loop on vertices and their cc id
#     #
#     for v, v_cc, v_rev, v_rev_cc in ccompcont:
#         #
#         # cc ids must be not None
#         #
#         if v_cc is None or v_rev_cc is None:
#             raise NotNoneCCompId(v, v_cc, v_rev, v_rev_cc)
#         #
#         # Undiscovered cc
#         #
#         if l_cc_trans[v_cc] is None:
#             l_cc_trans[v_cc] = n_new_cc
#             graph_cc: GraphsCnntedCompT
#             #
#             # Symmetric cc
#             #
#             if v_rev_cc != v_cc:
#                 graph_cc = SplitStrandsGraph()
#                 l_cc_trans[v_rev_cc] = -1
#             #
#             # Merged cc
#             #
#             else:
#                 graph_cc = MergedStrandsRevSymGraph()
#             __add_adj_graph_cc(graph, graph_cc, v)
#             graph_ccs.append(graph_cc)
#             n_new_cc += 1
#         #
#         # Already discovered cc
#         #
#         elif l_cc_trans[v_cc] < 0:  # type: ignore
#             # Reverse sym cc with arbitrarly v_rev_cc kept
#             __add_adj_graph_cc(
#                 graph, graph_ccs[l_cc_trans[v_rev_cc]], v_rev,  # type: ignore
#             )

#         else:
#             #
#             # v_cc kept
#             #
#             __add_adj_graph_cc(
#                 graph, graph_ccs[l_cc_trans[v_cc]], v,  # type: ignore
#             )
#             #
#             # Must add edges for the reverse bc reverse merged
#             #
#             if v_cc == v_rev_cc:
#                 __add_adj_graph_cc(
#                     graph, graph_ccs[l_cc_trans[v_rev_cc]],  # type: ignore
#                     v_rev,
#                 )

#     return graph_ccs


# def __add_adj_graph_cc(graph: RevSymGraph, graph_cc: GraphsCnntedCompT,
#                        vertex: IndOrT):
#     """Add `v` adjencies edges to `graph_cc`.

#     Parameters
#     ----------
#     graph : RevSymGraph
#         Reverse symmetric graph
#     graph_cc : SplitStrandsOverlapsgraph or MergedStrandsRevSymGraph
#         Connected component graph
#     vertex : IndOrT
#         Oriented vertex
#     """
#     vertices = graph.vertices()
#     edges = graph.edges()
#     new_vertices = graph_cc.vertices()
#     new_edges = graph_cc.edges()
#     if isinstance(new_vertices, SplitStrandsVertices):
#         v_new_ind = new_vertices.add(vertex[OR])
#     else:
#         v_new_ind = new_vertices.add()
#     # XXX: be sure to add attr before!
#     for attrname, attrvalue in vertices.attrs(vertex[IND]):
#         new_vertices.set_attr(v_new_ind, attrname, attrvalue)
#     for w, e_ind in edges.succs(vertex):
#         w_new =
#         graph_cc.add_overlap(
#             (v_new, w_new), **edges.attrs(((v, w), e_ind)),
#         )



# -*- coding=utf-8 -*-

"""0-based indices reads' identifiers container class module."""

from __future__ import annotations

from typing import Generator, Optional, cast

from revsymg.exceptions import NoReadId, NoVertexIndex
from revsymg.hash_ids.abstract_ids import AbstractIdsContainer
from revsymg.hash_ids.hashable_ids import HashableIdsContainer
from revsymg.lib.index_lib import FORWARD_INT, IND, IndexT, IndOrT, OrT


# ============================================================================ #
#                       INDICES 0 BASED READS IDENTIFIERS                      #
# ============================================================================ #
class IndicesIdsContainer(AbstractIdsContainer[IndexT]):
    """Reads identifier container for 0-based indentifiers."""

    def __init__(self):
        """The Initializer."""
        # Valid until the 0-based index property becomes false
        self._card_id: int = 0
        # Valid when the 0-based index property becomes false
        self._readsid_container: Optional[HashableIdsContainer] = None

    # -*- Identifiers -*-

    def id_to_indor(self, read_id: IndexT,
                    orientation: OrT = FORWARD_INT) -> IndOrT:
        """Return the oriented vertex associated to the id.

        Parameters
        ----------
        read_id : IdT
            Read's identifier
        orientation : OrT, default FORWARD_INT
            Read orientation

        Returns
        -------
        IndOrT
            Oriented vertex

        Raises
        ------
        NoReadId
            If there is no read's identifier in container
        """
        if self._readsid_container is None:
            if read_id < self._card_id:
                return read_id, orientation
            raise NoReadId(read_id)
        return self._readsid_container.id_to_indor(read_id, orientation)

    def indor_to_id(self, vertex: IndOrT) -> IndexT:
        """Return the id from vertices.

        Parameters
        ----------
        vertex : IndOrT
            Oriented vertex

        Returns
        -------
        IndexT
            The associated read's identifier

        Raises
        ------
        NoVertexIndex
            There is no read's identifier corresponding to given vertex
        """
        if self._readsid_container is None:
            if vertex[IND] >= self._card_id:
                raise NoVertexIndex(vertex[IND])
            return vertex[IND]
        return cast(IndexT, self._readsid_container.indor_to_id(vertex))

    # -*- Setter -*-

    def add(self, read_id: IndexT) -> IndexT:
        r"""Add an entry if identifier not yet recorded.

        Parameters
        ----------
        read_id : IndexT
            Read's identifier

        Returns
        -------
        IndexT
            Associated vertex index

        Notes
        -----
        It considers bijective-property respected (read's identifiers are
        0-base indices), thus it adds all read's identifiers between the
        :math:`max(r_{id} \in \mathcal{R_{id}}` and new :math:`r_{id}`

        If the property was violated, only new :math:`r_{id}` is added.
        """
        if self._readsid_container is None:
            if read_id < self._card_id:
                return read_id
            self._card_id += 1
            return self._card_id - 1
        return self._readsid_container.add(read_id)

    def pop(self, read_id: IndexT) -> IndexT:
        """Remove read's identifier.

        Parameters
        ----------
        read_id : int
            Read's identifier

        Returns
        -------
        IndexT
            Index corresponding to read's identifier

        Raises
        ------
        NoReadId
            If there is no read's identifier in container
        """
        if self._readsid_container is None:
            if read_id > self._card_id - 1:
                raise NoReadId(read_id)
            if read_id == self._card_id - 1:  # property not yet violated
                self._card_id -= 1
            elif read_id < self._card_id - 1:  # property violated
                self._readsid_container = HashableIdsContainer()
                for r_id in range(self._card_id):
                    if r_id != read_id:
                        self._readsid_container.add(r_id)
            return read_id
        return self._readsid_container.pop(read_id)

    # -*- Overloaded built-in -*-

    def __iter__(self) -> Generator[IndexT, None, None]:
        """Iterate on reads' identifier.

        Yields
        ------
        IndexT
            Reads' identifier
        """
        if self._readsid_container is None:
            yield from range(self._card_id)
        else:
            yield from (
                cast(IndexT, read_id) for read_id in self._readsid_container
            )

    def __contains__(self, read_id: IndexT) -> bool:
        """Return True if read's identifier in container.

        Parameters
        ----------
        read_id : IndexT
            Read's identifier

        Returns
        -------
        bool
            Is read's identifier in container?
        """
        if self._readsid_container is None:
            return read_id < self._card_id
        return read_id in self._readsid_container

    def __len__(self) -> int:
        """Return the number of recorded read's identifiers.

        Returns
        -------
        int
            Number of reads' identifiers
        """
        if self._readsid_container is None:
            return self._card_id
        return len(self._readsid_container)



# -*- coding=utf-8 -*-

"""Hashable reads' identifiers container class module."""

from __future__ import annotations

from collections.abc import Hashable
from typing import Generator

from revsymg.exceptions import NoReadId, NoVertexIndex
from revsymg.hash_ids.abstract_ids import AbstractIdsContainer
from revsymg.lib.index_lib import FORWARD_INT, IND, IndexT, IndOrT, OrT


# ============================================================================ #
#                          HASHABLE READS IDENTIFIERS                          #
# ============================================================================ #
class HashableIdsContainer(AbstractIdsContainer[Hashable]):
    """Reads identifier container for hashable identifiers."""

    def __init__(self):
        """The initializer."""
        # List vertex index to vertex id
        #   [v.index]: v.idr
        self._indices_ids: list[Hashable] = []
        # dictionary vertex id to vertex index
        #   v.idr -> v.index
        self._ids_indices: dict[Hashable, IndexT] = {}

    # -*- Identifier -*-

    def id_to_indor(self, read_id: Hashable,
                    orientation: OrT = FORWARD_INT) -> IndOrT:
        """Return the oriented vertex associated to the id.

        Parameters
        ----------
        read_id : Hashable
            Read's identifier
        orientation : OrT, default FORWARD_INT
            Read orientation

        Returns
        -------
        IndOrT
            Oriented vertex

        Raises
        ------
        NoReadId
            If there is no read's identifier in container
        """
        try:
            return self._ids_indices[read_id], orientation
        except KeyError as exc:
            raise NoReadId(read_id) from exc

    def indor_to_id(self, vertex: IndOrT) -> Hashable:
        """Return the id from vertices.

        Parameters
        ----------
        vertex : IndOrT
            Oriented vertex

        Returns
        -------
        Hashable
            The associated read's identifier

        Raises
        ------
        NoVertexIndex
            There is no read's identifier corresponding to given vertex
        """
        try:
            return self._indices_ids[vertex[IND]]
        except IndexError as exc:
            raise NoVertexIndex(vertex[IND]) from exc

    # -*- Setter -*-

    def add(self, read_id: Hashable) -> IndexT:
        """Add an entry if identifier not yet recorded else corresponding index.

        Parameters
        ----------
        read_id : Hasable
            Read's identifier

        Returns
        -------
        IndexT
            Associated vertex index
        """
        try:
            return self._ids_indices[read_id]
        except KeyError:
            self._ids_indices[read_id] = len(self._indices_ids)
            self._indices_ids.append(read_id)
        return self._ids_indices[read_id]

    def pop(self, read_id: Hashable) -> IndexT:
        """Remove read's identifier.

        Parameters
        ----------
        read_id : Hashable
            Read's identifier

        Returns
        -------
        IndexT
            Index corresponding to read's identifier

        Raises
        ------
        NoReadId
            If there is no read's identifier in container
        """
        try:
            index = self._ids_indices.pop(read_id)
        except KeyError as exc:
            raise NoReadId(read_id) from exc
        self._indices_ids = (
            self._indices_ids[:index] + self._indices_ids[index + 1:]
        )
        for post_id in self._indices_ids[index:]:
            self._ids_indices[post_id] -= 1
        return index

    # -*- Overloaded built-in -*-

    def __iter__(self) -> Generator[Hashable, None, None]:
        """Iterate on reads' identifier.

        Yields
        ------
        Hashable
            Reads' identifier
        """
        yield from self._ids_indices.keys()

    def __contains__(self, read_id: Hashable) -> bool:
        """Return True if read's identifier in container.

        Parameters
        ----------
        read_id : Hashable
            Read's identifier

        Returns
        -------
        bool
            Is read's identifier in container?
        """
        return read_id in self._ids_indices

    def __len__(self) -> int:
        """Return the number of recorded read's identifiers.

        Returns
        -------
        int
            Number of reads' identifiers
        """
        return len(self._indices_ids)



# -*- coding=utf-8 -*-

"""Attributes container class module."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Iterator

from revsymg.exceptions import WrongAttributeType, _NoAttribute, _NoKey
from revsymg.lib.index_lib import IndexT


# ============================================================================ #
#                           ATTRIBUTES CONCRETE CLASS                          #
# ============================================================================ #
class AttributesContainer():
    """Attributes container class."""

    def __init__(self):
        """The Initializer."""
        self._card_keys: int = 0
        self._attrname_index_default: dict[str: tuple[IndexT, Any]] = {}
        self._lists_values: list[list[Any]] = []

    def card_properties(self) -> int:
        """Return the number of attributes fields.

        Returns
        -------
        int
            Number of attributes fields
        """
        return len(self._lists_values)

    def names(self) -> Iterable[str]:
        """Return the attributes name.

        Returns
        -------
        iterable of str
            Attributes names
        """
        return self._attrname_index_default

    def has_attrname(self, attrname: str) -> bool:
        """Return True if there is a property named attrname.

        Parameters
        ----------
        attrname : str
            Property's name

        Returns
        -------
        bool
            Is attrname in property names?
        """
        return attrname in self._attrname_index_default

    def get(self, key: IndexT, attrname: str) -> Any:
        """Return the value associated to the `key` attribute `attrname`.

        Parameters
        ----------
        key : IndexT
            Object index
        attrname : str
            Attribute name

        Returns
        -------
        Any
            Attribute value

        Raises
        ------
        _NoAttribute
            No attribute
        _NoKey
            Key does not exist
        """
        try:
            attr_index, _ = self._attrname_index_default[attrname]
        except KeyError as key_error:
            raise _NoAttribute(attrname) from key_error
        try:
            return self._lists_values[attr_index][key]
        except IndexError as index_error:
            raise _NoKey(key) from index_error

    def get_all(self, key: IndexT) -> Iterator[tuple[str, Any]]:
        """Iterate over attribute name and values for the key.

        Parameters
        ----------
        key : IndexT
            Object index

        Yields
        ------
        str, Any
            Attribute name, and all values

        Raises
        ------
        _NoKey
            Key does not exist
        """
        if key >= self._card_keys:
            raise _NoKey(key)
        for attrname, (attr_index, _) in self._attrname_index_default.items():
            yield attrname, self._lists_values[attr_index][key]

    def new_attr(self, attrname: str, default: Any):
        """Add an attribute entry.

        Parameters
        ----------
        attrname : str
            Attribute's name
        default : Any
            Default keys' attribute value
        """
        if attrname in self._attrname_index_default:
            attr_index, _ = self._attrname_index_default[attrname]
        else:
            attr_index = len(self._attrname_index_default)
            self._lists_values.append([])
        self._attrname_index_default[attrname] = (attr_index, default)
        self._lists_values[attr_index] = [default] * self._card_keys

    def add_keys(self, number: int = 1):
        """Add number of keys.

        Parameters
        ----------
        number : int, default 1
            Number of keys to add
        """
        self._card_keys += number
        for attr_index, default in self._attrname_index_default.values():
            self._lists_values[attr_index] += [default] * number

    def set_attr(self, key: IndexT, attrname: str, attrvalue: Any):
        """Set the values for the attributes, else create them.

        Parameters
        ----------
        key : IndexT
            Object index
        attrname : str
            Attribute's name
        attrvalue : Any
            Attribute's value

        Raises
        ------
        _NoAttribute
            There is one attribute's name that was not added before
        WrongAttributeType
            Given value does not correspond to attribute's type
        _NoKey
            Key does not exist
        """
        try:
            attr_index, default = self._attrname_index_default[attrname]
        except KeyError as key_error:
            raise _NoAttribute(attrname) from key_error
        if not isinstance(attrvalue, type(default)):
            raise WrongAttributeType(attrvalue, type(default))
        try:
            self._lists_values[attr_index][key] = attrvalue
        except IndexError as index_error:
            raise _NoKey(key) from index_error

    def delete_key(self, key: IndexT):
        """Delete an entire key.

        Parameters
        ----------
        key : IndexT
            Key to delete

        Raises
        ------
        _NoKey
            Key does not exist

        Warning
        -------
        This operation may invalidate keys' index. In fact, deleting one key
        implies shifting all next indices. Because of this, keys should always
        be removed in decreasing index order:

        >>> for key in sorted(keys_to_del, reverse=True):
        >>>     attr_container.delete_key(key)
        """
        if key >= self._card_keys:
            raise _NoKey(key)
        for attr_index, values in enumerate(self._lists_values):
            self._lists_values[attr_index] = values[:key] + values[key + 1:]
        self._card_keys -= 1

    def delete_keys(self, keys: Iterable[IndexT]):
        """Delete entire keys.

        Parameters
        ----------
        keys : iterable of IndexT
            Keys to delete

        Raises
        ------
        _NoKey
            Key does not exist

        Note
        ----
        At the oposite of :meth:`~revsymg.AttributesContainer.delete_key`
        method, it is not necessary to sort `keys`. This is done internally.
        """
        for key in sorted(keys, reverse=True):
            if key >= self._card_keys:
                raise _NoKey(key)
            for attr_index, values in enumerate(self._lists_values):
                self._lists_values[attr_index] = values[:key] + values[key + 1:]
            self._card_keys -= 1



# # -*- coding=utf-8 -*-

"""Transitive edges removal algo."""


from typing import Literal

from revsymg.graphs.revsym_graph import RevSymGraph
from revsymg.lib.index_lib import IND, IndOrT


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# REFACTOR: move these constants?
FUZZ: Literal[10] = 10
"""Constant for upper-bound."""

READ_LEN_STR = 'read_length'
"""Vertex attribute name coresponding to read length."""

OV_LEN_STR = 'overlap_length'
"""Edge attribute name corresponding to overlap length."""

_VACANT = 0
_INPLAY = 1
_ELIMINATED = 2


# ============================================================================ #
#                                 MAIN FUNCTION                                #
# ============================================================================ #
def transitive_reduction(ovg: RevSymGraph, fuzz: int = FUZZ,
                         read_len_str: str = READ_LEN_STR,
                         ov_len_str: str = OV_LEN_STR):
    """Remove transitive edges.

    Parameters
    ----------
    ovg : RevSymGraph
        Overlaps graph
    fuzz : int, default FUZZ
        Constant for upper-bound
    read_len_str : str, default READ_LEN_STR
        Vertex attribute name coresponding to read length
    ov_len_str : str, default OV_LEN_STR
        Edge attribute name corresponding to overlap length

    Warning
    -------
    For each pair oriented vertices :math:`u` and :math:`v` there is
    **at most one** overlap.
    """
    # NOTE: do proof concerning non-colision in indices
    #   - especially concerning tandem and return of self
    l_mark = [_VACANT] * len(ovg.vertices())

    for v in ovg.vertices():

        lt_edges = __one_top_neighbors(ovg, v, read_len_str, ov_len_str)

        longest = lt_edges[-1][1] + fuzz if lt_edges else 0
        # -------------------------------------------------------------------- #
        # Detect transitive edges
        # -------------------------------------------------------------------- #
        for w, _ in lt_edges:
            l_mark[w[IND]] = _INPLAY

        for w, w_suffix in lt_edges:

            lt_2_edges = __two_top_neighbors(
                ovg, w, l_mark, read_len_str, ov_len_str,
            )

            for k, (x, x_suffix) in enumerate(lt_2_edges):

                if (l_mark[w[IND]] == _INPLAY
                        and w_suffix + x_suffix <= longest):
                    l_mark[x[IND]] = _ELIMINATED
                # pylint: disable=compare-to-zero
                if x_suffix < fuzz or k == 0:
                    l_mark[x[IND]] = _ELIMINATED
        # -------------------------------------------------------------------- #
        # Remove transitive edges
        # -------------------------------------------------------------------- #
        for w, e_ind in ovg.edges().succs(v):
            if l_mark[w[IND]] == _ELIMINATED:
                # XXX: proof rev delete
                ovg.edges().delete(v, w, e_ind)
            l_mark[w[IND]] = _VACANT


def __one_top_neighbors(ovg: RevSymGraph,
                        v: IndOrT, read_len_str: str,
                        ov_len_str: str) -> list[tuple[IndOrT, int]]:
    """Return the one-top neighbour of `v` sorted by suffix length.

    Parameters
    ----------
    ovg : RevSymGraph
        Overlaps graph
    v : IndOrT
        Oriented vertex
    read_len_str : str
        Attribute name for read's length
    ov_len_str : str
        Attribute name for overlap's length

    Returns
    -------
    list of IndOrT, int
        List of sorted successors and their suffix length
    """
    lt_edges: list[tuple[IndOrT, int]] = []

    for w, e_ind in ovg.edges().succs(v):

        w_suffix = max(
            0,
            ovg.vertices().attr(w[IND], read_len_str)
            - ovg.edges().attr(e_ind, ov_len_str),
        )
        lt_edges.append((w, w_suffix))

    return sorted(lt_edges, key=lambda t: t[1])


def __two_top_neighbors(ovg: RevSymGraph, w: IndOrT,
                        l_mark: list[int], read_len_str: str,
                        ov_len_str: str) -> list[tuple[IndOrT, int]]:
    """Return the one-top neighbors of `w` sorted by suffix length.

    Parameters
    ----------
    ovg : RevSymGraph
        Overlaps graph
    w : IndOrT
        Oriented vertex
    l_mark : list of int
        Vertices states
    read_len_str : str
        Attribute name for read's length
    ov_len_str : str
        Attribute name for overlap's length

    Returns
    -------
    list of IndOrT, int
        List of sorted successors and their suffix length
    """
    lt_edges: list[tuple[IndOrT, int]] = []

    for x, e_ind in ovg.edges().succs(w):
        if l_mark[x[IND]] == _INPLAY:
            x_suffix = max(
                0,
                ovg.vertices().attr(x[IND], read_len_str)
                - ovg.edges().attr(e_ind, ov_len_str),
            )
            lt_edges.append((x, x_suffix))

    return sorted(lt_edges, key=lambda t: t[1])



# -*- coding=utf-8 -*-

"""Types, constants and functions for indices module."""

from __future__ import annotations

from typing import Literal


# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
# REFACTOR: separate & rename files to be more precise
#   - Base
#   - Reads
#   - Overlaps
# ---------------------------------------------------------------------------- #
#                                     Base                                     #
# ---------------------------------------------------------------------------- #
IndexT = int
"""Indices type."""

OrT = Literal[0, 1]
"""Orientations type."""
# ---------------------------------------------------------------------------- #
#                                   Oriented                                   #
# ---------------------------------------------------------------------------- #
IndOrT = tuple[IndexT, OrT]
"""Vertices type."""

EIndOrT = tuple[IndOrT, IndOrT]
"""Edges type."""

# DOC: change doc and remove EIndorT
EIndOrIndT = tuple[IndOrT, IndOrT, IndexT]
"""Edges with edge index type."""

IndOrIndT = tuple[IndOrT, IndexT]
"""Adjacent vertex with edge index type."""
# ---------------------------------------------------------------------------- #
#                                  Unoriented                                  #
# ---------------------------------------------------------------------------- #
# DOC: do doc EIndT
EIndT = tuple[IndexT, IndexT]
"""Edges of indices type."""

# DOC: do doc EIndIndT
EIndIndT = tuple[IndexT, IndexT, IndexT]
"""Edges of indices with edge index type."""

# DOC: do doc IndIndT
IndIndT = tuple[IndexT, IndexT]
"""Adjacent index vertex with edge index type."""

# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                             Fragment Orientations                            #
# ---------------------------------------------------------------------------- #
FORWARD_INT: OrT = 0
"""Forward orientation value.

type: :data:`OrT`
"""

REVERSE_INT: OrT = 1
"""Reverse orientation value.

type: :data:`OrT`
"""
#
# The list that gives the reverse value of a given orientation.
#
ORIENT_REV: tuple[OrT, OrT] = (REVERSE_INT, FORWARD_INT)
""":data:`OrT` reverse list.

type: :class:`tuple` (:data:`OrT`, :data:`OrT`)
"""
#
# For graph when ask oriented edges
#
FWDOR_REVOR: tuple[tuple[OrT, OrT], tuple[OrT, OrT]] = (
    (FORWARD_INT, REVERSE_INT),
    ORIENT_REV,
)
""":data:`OrT` relative orientation.

type: :class:`tuple` ((:data:`OrT`, :data:`OrT`), (:data:`OrT`, :data:`OrT`))
"""
# ---------------------------------------------------------------------------- #
#                                Oriented Vertex                               #
# ---------------------------------------------------------------------------- #
IND: Literal[0] = 0
"""The index for :data:`IndOrT`'s index value."""
OR: Literal[1] = 1
"""The index for :data:`IndOrT`'s orientation value."""
# ---------------------------------------------------------------------------- #
#                                Adjancies Table                               #
# ---------------------------------------------------------------------------- #
PRED_IND: Literal[0] = 0
"""The index for predecessors of forwards in adjancies table"""
SUCC_IND: Literal[1] = 1
"""The index for successors of forwards in adjancies table"""
# ---------------------------------------------------------------------------- #
#                                 Adjancy Type                                 #
# ---------------------------------------------------------------------------- #
ADJ_INDOR: Literal[0] = 0
"""The index for :data:`IndOrT` value in :data:`IndOrIndT`."""
ADJ_IND: Literal[1] = 1
"""The index for :data:`IndexT` value in :data:`IndOrIndT`."""
# ---------------------------------------------------------------------------- #
#                             Edges And Edge Index                             #
# ---------------------------------------------------------------------------- #
U_INDOR: Literal[0] = 0
"""The index of first vertex in the edge."""

V_INDOR: Literal[1] = 1
"""The index of second vertex in the edge."""

E_IND: Literal[2] = 2
"""The index of edge's index in the edge.

For both :data:`EIndOrIndT` and :data:`EIndIndT` it corresponds to
:data:`IndexT` value.
"""
# ---------------------------------------------------------------------------- #
#                                 Miscellaneous                                #
# ---------------------------------------------------------------------------- #
MULT_IID: Literal[2] = 2
"""Index multiplier."""


# ============================================================================ #
#                                   FONCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                    IndOrT                                    #
# ---------------------------------------------------------------------------- #
def is_forward(vertex: IndOrT) -> bool:
    """Return True if the `vertex` is considered as forward.

    Parameters
    ----------
    vertex : IndOrT
        A vertex

    Returns
    -------
    bool
        True if vertex's orientation is forward
    """
    return vertex[OR] == FORWARD_INT


def is_reverse(vertex: IndOrT) -> bool:
    """Return True if the `vertex` is considered as reverse.

    Parameters
    ----------
    vertex : IndOrT
        Oriented vertex

    Returns
    -------
    bool
        True is vertex is in reverse orientation
    """
    return vertex[OR] == REVERSE_INT


def rev_vertex(vertex: IndOrT) -> IndOrT:
    """Return the reverse vertex.

    Parameters
    ----------
    vertex : IndOrT
        Oriented vertex

    Returns
    -------
    IndOrT
        The reversed of vertex
    """
    return vertex[IND], ORIENT_REV[vertex[OR]]


# ---------------------------------------------------------------------------- #
#                                    EIndOrT                                   #
# ---------------------------------------------------------------------------- #
def rev_edge(first_vertex: IndOrT, second_vertex: IndOrT) -> EIndOrT:
    """Return the reverse edge.

    Parameters
    ----------
    first_vertex : IndOrT
        First oriented vertex in the edge
    second_vertex : IndOrT
        Second oriented vertex in the edge

    Returns
    -------
    EIndOrT
        The reversed of edge
    """
    return (
        (second_vertex[IND], ORIENT_REV[second_vertex[OR]]),
        (first_vertex[IND], ORIENT_REV[first_vertex[OR]]),
    )


def is_canonical(first_vertex: IndOrT, second_vertex: IndOrT) -> bool:
    r"""Xor boolean function between edge and its reverse.

    Parameters
    ----------
    first_vertex : IndOrT
        First oriented vertex in the edge
    second_vertex : IndOrT
        Second oriented vertex in the edge

    Returns
    -------
    bool
        True if :math:`\texttt{is_canonical}(\overline{edge})` is False,
        else False

    Notes
    -----
    :math:`\texttt{is_canonical}` is defined such as
    :math:`\texttt{is_canonical}(edge)
    \veebar \texttt{is_canonical}(\overline{edge})`
    """
    return (
        first_vertex[OR] == FORWARD_INT
        and (
            second_vertex[OR] == FORWARD_INT
            or first_vertex[IND] < second_vertex[IND]
        )
        or (
            second_vertex[OR] == FORWARD_INT
            and first_vertex[IND] < second_vertex[IND]
        )
    )


def eindor_orientation(first_vertex: IndOrT, second_vertex: IndOrT) -> OrT:
    """Return the edge orientation.

    Parameters
    ----------
    first_vertex : IndOrT
        First oriented vertex in the edge
    second_vertex : IndOrT
        Second oriented vertex in the edge

    Returns
    -------
    OrT
        Edge orientation

    Notes
    -----
    This is pure definition.
    Note that in case of self loop, return value is
    :const:`~revsymg.lib.index_lib.FORWARD_INT`
    """
    if first_vertex[IND] < second_vertex[IND]:
        return FORWARD_INT
    if first_vertex[IND] > second_vertex[IND]:
        return REVERSE_INT
    if first_vertex[OR] < second_vertex[OR]:
        return FORWARD_INT
    if first_vertex[OR] > second_vertex[OR]:
        return REVERSE_INT
    return first_vertex[OR]
